
/******************************************************************************
* MODULE     : dense_pivot.hpp
* DESCRIPTION: Pivoting for dense matrices
* COPYRIGHT  : (C) 2005  Olivier Ruatta and Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_DENSE_PIVOT_HPP
#define __MMX_DENSE_PIVOT_HPP
#include "dense_matrix.hpp"

namespace mmx {
#define TMPL template<typename C, typename V>
#define Matrix matrix<C,dense_matrix<V> >
#define Matrix_rep matrix_rep<C,dense_matrix<V> >

/******************************************************************************
* Pivoting
******************************************************************************/

// NOTE for Olivier:
// please add or replace by corrected versions of the previous algorithms

/* Brutal method */
TMPL void
naive_pivoting (Matrix& m,
		const nat& ri, const nat& ci,
		const nat& rf, const nat& cf)
{
  assert ((ri < rows(m)) && (ci < cols(m)) &&
	  (rf <= rows(m)) && (cf <= cols(m)));

  // search pivot
  nat i= ci, j= ri;
  for (j= ci; j<cf; j++) {
    for (i= ri; i<rf; i++)
      if (m (i, j) != 0) break;
    if (i<rf) break;
  }

  if (i<rf && j<cf) {
    // put pivot on the top and the left
    if (i != ri) in_place_swap_row (m, ri, i);
    if (j != ci) in_place_swap_col (m, ci, j);
    // operations are made on the rows
    for (nat index = 1; index < rf-ri; index++) 
      in_place_rows_linsub (m, ri + index, m (ri, ci), ri, m (ri+index, ci));
    // call the pivoting on the non triangular sub-matrix
    if (ri+1 < rf && ci+1 < cf) 
      naive_pivoting (m, ri+1, ci+1, rf, cf);
  }
}

/* Brutal method */
TMPL void
naive_extended_pivoting (Matrix& m, Matrix& k,
			 nat ri, nat ci,
			 nat rf, nat cf)
{
  assert ((ri < rows(m)) && (ci < cols(m)) &&
	  (rf <= rows(m)) && (cf <= cols(m)));

  // search pivot
  nat i= ci, j= ri;
  for (j= ci; j<cf; j++) {
    for (i= ri; i<rf; i++)
      if (m (i, j) != 0) break;
    if (i<rf) break;
  }

  if (i<rf && j<cf) {
    // put pivot on the top and the left
    if (i != ri) {
      in_place_swap_row (k, ri, i);
      in_place_swap_row (m, ri, i);
    }
    ci= j;
    // operations are made on the rows
    for (nat index = 1; index < rf-ri; index++) {
      in_place_rows_linsub (k, ri + index, m (ri, ci), ri, m (ri+index, ci));
      in_place_rows_linsub (m, ri + index, m (ri, ci), ri, m (ri+index, ci));
    }
    // call the pivoting on the non triangular sub-matrix
    if (ri+1 < rf && ci+1 < cf) 
      naive_extended_pivoting (m, k, ri+1, ci+1, rf, cf);
  }
}

/******************************************************************************
* Application
******************************************************************************/

/* Brutal rank computation */
TMPL nat
rank (const Matrix& m) {
  assert (is_matricial (m));
  Matrix c (copy (m));
  naive_pivoting (c, 0, 0, rows(c), cols(c));
  nat i;
  for (i=0; i < min (rows(c), cols(c)) && c(i,i) != 0; i++);
  return i;
}

/* Brutal rank computation */
TMPL Matrix
left_kernel (const Matrix& m) {
  assert (is_matricial (m));
  Matrix c (copy (m));
  Matrix k (C(1), rows(m), rows(m));
  naive_extended_pivoting (c, k, 0, 0, rows(c), cols(c));
  nat i, j;
  for (i=0; i<rows(c); i++)
    if (segment_eq (tab(c)+i*cols(c), C(0), cols(c))) break;
  k= submatrix (k, i, rows(m), 0, rows(m));
  for (i=0; i<rows(k); i++)
    for (j=0; j<cols(k); j++)
      if (k(i,j) != 0) return k / k(i,j);
  return k;
}

/* Brutal determinant computation */
TMPL C
det (Matrix& m) {
  assert (is_matricial (m));
  assert (cols(m) == rows(m));
  Matrix c (copy (m));
  naive_pivoting (c, 0, 0, rows(c)-1, cols(c)-1);
  return c (rows(c)-1, rows(c)-1);
}

}//end of namespace
#undef TMPL
#undef Matrix
#undef Matrix_rep
#endif //__MMX_DENSE_PIVOT
