
/******************************************************************************
* MODULE     : series.hpp
* DESCRIPTION: Univariate power series
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_SERIES_HPP
#define __MMX_SERIES_HPP
#include "dense_polynomial.hpp"
#include "list.hpp"
#include "table.hpp"
namespace mmx {
#define TMPL_DEF template<typename C, typename V=naive_multiply>
#define TMPL template<typename C, typename V>
#define Series series<C,V>
#define Series_rep series_rep<C,V>
#define Implicit_series_rep implicit_series_rep<C,V>
TMPL class series;
TMPL generic flatten (const Series& f, const generic& z);

/******************************************************************************
* Global variables
******************************************************************************/

extern nat  mmx_series_precision;       // current series precision
extern nat  mmx_series_cancel_order;    // order for zero testing
extern bool mmx_series_formula_output;  // print formulas vs. expansions

inline nat set_series_precision (nat prec) {
  nat old= mmx_series_precision;
  mmx_series_precision= prec;
  return old; }
inline void reset_series_precision (nat prec) {
  mmx_series_precision= prec; }
inline nat get_series_precision () {
  return mmx_series_precision; }

inline nat set_series_cancel_order (nat prec) {
  nat old= mmx_series_cancel_order;
  mmx_series_cancel_order= prec;
  return old; }
inline void reset_series_cancel_order (nat prec) {
  mmx_series_cancel_order= prec; }
inline nat get_series_cancel_order () {
  return mmx_series_cancel_order; }

inline bool set_series_formula_output (bool mode) {
  bool old= mmx_series_formula_output;
  mmx_series_formula_output= mode;
  return old; }
inline void reset_series_formula_output (bool mode) {
  mmx_series_formula_output= mode; }
inline bool get_series_formula_output () {
  return mmx_series_formula_output; }

/******************************************************************************
* Dense series
******************************************************************************/

TMPL_DEF
class series_rep REP_STRUCT {
public:   // should be protected
  C*  a;  // coefficients
  nat n;  // number of computed coefficients
  nat l;  // number of allocated coefficients
          // a[n],...,a[l-1] may be used by relaxed computations

  inline series_rep (): a (new C[0]), n (0), l (0) {}
  inline virtual ~series_rep () { delete[] a; }
  inline Series me () const;
  virtual C next () = 0;
  virtual generic expression (const generic& z) const = 0;

public:
  virtual void Set_precision (nat l2);
  virtual void Increase_precision (nat l2=0);
  friend class Series;
};

TMPL_DEF
class series {
INDIRECT_PROTO_2 (series, series_rep, C, V)
public:
  series ();
  template<typename T> series (const T& c);
  template<typename T> series (const T& c, nat deg);
  template<typename T> series (const polynomial<T>& P);
  template<typename T> series (const series<T,V>& f);
  series (const iterator<C>& it, const string& name= "explicit");
  series (C (*coeffs) (nat), const string& name= "explicit");
  C operator [] (nat n) const;
  const C* operator () (nat start, nat end) const;
};
INDIRECT_IMPL_2 (series, series_rep, typename C, C, typename V, V)

STYPE_TO_TYPE(TMPL,scalar_type,Series,C);

/******************************************************************************
* Implicit series
******************************************************************************/

TMPL
class implicit_series_rep: public Series_rep {
public:
  Series eq;
public:
  inline implicit_series_rep () {}
  virtual Series initialize () = 0;
  inline C next () { return eq[this->n]; }
  C& initial (nat n2) {
    if (n2>=this->n) { this->n = n2+1; Set_precision (this->n); }
    return this->a[n2]; }
  virtual void Increase_precision (nat l) {
    Series_rep::Increase_precision (l);
    increase_precision (eq, l); }
};

TMPL
class implicit_container_series_rep: public Series_rep {
  Series f;
public:
  implicit_container_series_rep (Series f2): f (f2) {
    Implicit_series_rep* rep= (Implicit_series_rep*) f.operator -> ();
    rep->eq= rep->initialize (); }
  ~implicit_container_series_rep () {
    Implicit_series_rep* rep= (Implicit_series_rep*) f.operator -> ();
    rep->eq= Series (); }
  C next () { return f[this->n]; }
  generic expression (const generic& z) const { return flatten (f, z); }
  virtual void Increase_precision (nat l) {
    Series_rep::Increase_precision (l);
    increase_precision (f, l); }
};

TMPL Series
implicit (const Series& f) {
  return (Series_rep*) new implicit_container_series_rep<C,V> (f);
}

/******************************************************************************
* Routines on dense series
******************************************************************************/

TMPL inline generic var (const Series& f) {
  (void) f; return "z";
}

TMPL inline Series
Series_rep::me () const {
  return Series (this, true);
}

TMPL void
Series_rep::Set_precision (nat l2) {
  if (l2 <= l) return;
  C* b= new C[l2];
  jet_copy (b, a, l);
  jet_clear (b+l, l2-l);
  delete[] a;
  a= b;
  l= l2;
}

TMPL void
Series_rep::Increase_precision (nat l2) {
  l2= max (l2, l);
  Set_precision (max ((nat) 1, l2 << 1));
}

TMPL void
increase_precision (const Series& f, nat l) {
  if (f->l < l)
    (const_cast<Series_rep*> (f.operator -> ())) -> Increase_precision (l);
}

TMPL C
Series::operator [] (nat n) const {
  if (n <  rep->n) return rep->a[n];
  if (n >= rep->l) rep->Increase_precision (n+1);
  while (rep->n <= n) {
    rep->a[rep->n]= rep->next ();
    //cout << "{" << rep->n << ";" << this << "}";
    rep->n++;
  }
  // cout << (*this) << " [" << n << "] -> " << rep->a[n] << "\n";
  return rep->a[n];
}

TMPL const C*
Series::operator () (nat start, nat end) const {
  if (end <= rep->n) return rep->a + start;
  if (end >= rep->l) rep->Increase_precision (end);
  while (rep->n < end) {
    rep->a[rep->n]= rep->next ();
    //cout << "{" << rep->n << ";" << this << "}";
    rep->n++;
  }
  return rep->a + start;
}

/******************************************************************************
* Series iterator
******************************************************************************/

TMPL
class series_iterator_rep: public iterator_rep<C> {
  Series f;  // the series to traverse
  nat i;     // current position
protected:
  bool is_busy () { return true; }
  void advance () { i++; }
  C current () { return f[i]; }
public:
  series_iterator_rep (const Series& f2): f(f2), i(0) {}
};

TMPL iterator<C>
iterate (const Series& f) {
  return iterator<C> (new series_iterator_rep<C,V> (f));
}

/******************************************************************************
* Hashing and printing series
******************************************************************************/

TMPL nat
hash (const Series& s) {
  register nat i, h= 7531;
  for (i=0; i<4; i++)
    h= (h<<1) ^ (h<<5) ^ (h>>27) ^ hash (s[i]);
  return h;
}

TMPL generic
flatten (const Series& f, const generic& z) {
  if (get_series_formula_output ()) return f->expression (z);
  generic s= 0;
  nat order= get_series_precision ();
  for (nat i=0; i<order; i++)
    s= s + flatten (f[i]) * pow (z, i);
  s= s + generic ("O") (pow (z, generic (order)));
  return s;
}

TMPL generic
flatten (const Series& f) {
  return flatten (f, var (f));
}

PRINT_SUGAR(TMPL,Series)

/******************************************************************************
* Comparisons
******************************************************************************/

TMPL int
sign (const Series& f) {
  for (nat n=0; n<get_series_cancel_order(); n++) {
    int sgn= sign (f[n]);
    if (sgn != 0) return sgn;
  }
  return 0;
}

TMPL int
compare (const Series& f1, const Series& f2) {
  for (nat n=0; n<get_series_cancel_order(); n++) {
    int sgn= compare (f1[n], f2[n]);
    if (sgn != 0) return sgn;
  }
  return 0;
}

TMPL int
unequal (const Series& f1, const Series& f2) {
  for (nat n=0; n<get_series_cancel_order(); n++)
    if (f1[n] != f2[n]) return 1;
  return 0;
}

COMPARE_SUGAR(TMPL,Series);

/******************************************************************************
* Explicit series
******************************************************************************/

TMPL
class polynomial_series_rep: public Series_rep {
  polynomial<C> P;
public:
  polynomial_series_rep (const polynomial<C>& P2): P (P2) {}
  C next () { return this->n < dim(P)? P[this->n]: 0; }
  generic expression (const generic& z) const { return flatten (P, z); }
};

TMPL
Series::series () {
  rep= new polynomial_series_rep<C,V> (polynomial<C> (C(0), 0));
}

TMPL template<typename T>
Series::series (const T& c) {
  rep= new polynomial_series_rep<C,V> (polynomial<C> (C(c), 0));
}

TMPL template<typename T>
Series::series (const T& c, nat deg) {
  rep= new polynomial_series_rep<C,V> (polynomial<C> (C(c), deg));
}

TMPL template<typename T>
Series::series (const polynomial<T>& P) {
  rep= new polynomial_series_rep<C,V> (polynomial<C> (P));
}

template<typename C, typename V, typename K>
class cast_series_rep : public Series_rep {
  series<K,V> f;
public:
  cast_series_rep (const series<K,V>& f2): f (f2) {}
  C next () { return C (f[this->n]); }
  generic expression (const generic& z) const { return flatten (f, z); }
};

TMPL template<typename T>
Series::series (const series<T,V>& f) {
  rep= new cast_series_rep<C,V,T> (f);
}

TMPL
class iterator_series_rep : public Series_rep {
  iterator<C> it;
  string name;
public:
  iterator_series_rep (const iterator<C>& it2, const string& name2):
    it (it2), name (name2) {}
  C next () {
    if (busy (it)) {
      C c= *it; ++it; return c; }
    else return C(0); }
  generic expression (const generic& z) const {
    return flatten (name) (z); }
};

TMPL
Series::series (const iterator<C>& it, const string& name) {
  rep= new iterator_series_rep<C,V> (it, name);
}

/******************************************************************************
* Abstract operations on series
******************************************************************************/

template<typename Op,typename C,typename V>
class unary_series_rep : public Series_rep {
protected:
  Series f;
public:
  inline unary_series_rep (const Series& f2): f(f2) {}
  generic expression (const generic& z) const {
    return Op::op (flatten (f, z)); }
  virtual void Increase_precision (nat l) {
    Series_rep::Increase_precision (l);
    increase_precision (f, l); }
  virtual C next () { return Op::op (f[this->n]); }
};

template<typename Op,typename C,typename V> inline Series
unary_series (const Series& f) {
  return (Series_rep*) new unary_series_rep<Op,C,V> (f);
}

template<typename Op,typename C,typename V>
class unary_implicit_series_rep : public Implicit_series_rep {
protected:
  Series f;
  bool with_init;
  C c;
public:
  unary_implicit_series_rep (const Series& f2):
    f (f2), with_init (false) {}
  unary_implicit_series_rep (const Series& f2, const C& c2):
    f (f2), with_init (true), c (c2) {}
  generic expression (const generic& z) const {
    return Op::op (flatten (f, z)); }
  virtual void Increase_precision (nat l) {
    Implicit_series_rep::Increase_precision (l);
    increase_precision (f, l); }
  Series initialize () {
    if (with_init) this->initial (0)= c;
    else {
      assert (Op::nr_init () <= 1);
      if (Op::nr_init () == 1)
	this->initial (0)= Op::op (f[0]);
    }
    return Op::def (Series (this->me ()), f); }
};

template<typename Op,typename C,typename V> inline Series
unary_implicit_series (const Series& f) {
  Series_rep* rep= new unary_implicit_series_rep<Op,C,V> (f);
  return implicit (Series (rep));
}

template<typename Op,typename C,typename V> inline Series
unary_implicit_series (const Series& f, const C& c) {
  Series_rep* rep= new unary_implicit_series_rep<Op,C,V> (f, c);
  return implicit (Series (rep));
}

template<typename Op,typename C,typename V>
class binary_series_rep : public Series_rep {
protected:
  Series f, g;
public:
  inline binary_series_rep (const Series& f2, const Series& g2):
    f(f2), g (g2) {}
  generic expression (const generic& z) const {
    return Op::op (flatten (f, z), flatten (g, z)); }
  virtual void Increase_precision (nat l) {
    Series_rep::Increase_precision (l);
    increase_precision (f, l);
    increase_precision (g, l); }
  virtual C next () { return Op::op (f[this->n], g[this->n]); }
};

template<typename Op,typename C,typename V> inline Series
binary_series (const Series& f, const Series& g) {
  return (Series_rep*) new binary_series_rep<Op,C,V> (f, g);
}

template<typename Op,typename C,typename V>
class binary_implicit_series_rep : public Implicit_series_rep {
protected:
  Series f, g;
public:
  binary_implicit_series_rep (const Series& f2, const Series& g2):
    f(f2), g(g2) {}
  generic expression (const generic& z) const {
    return Op::op (flatten (f, z), flatten (g, z)); }
  virtual void Increase_precision (nat l) {
    Implicit_series_rep::Increase_precision (l);
    increase_precision (f, l);
    increase_precision (g, l); }
  Series initialize () {
    assert (Op::nr_init () <= 1);
    if (Op::nr_init () == 1) this->initial (0)= Op::op (f[0], g[0]);
    return Op::def (Series (this->me ()), f, g);
  }
};

template<typename Op,typename C,typename V> inline Series
binary_implicit_series (const Series& f, const Series& g) {
  Series_rep* rep= new binary_implicit_series_rep<Op,C,V> (f, g);
  return implicit (Series (rep));
}

template<typename Op,typename C,typename V,typename X>
class extrary_series_rep : public Series_rep {
protected:
  Series f;
  X x;
public:
  inline extrary_series_rep (const Series& f2, const X& x2):
    f(f2), x (x2) {}
  generic expression (const generic& z) const {
    return Op::op (flatten (f, z), flatten (x)); }
  virtual void Increase_precision (nat l) {
    Series_rep::Increase_precision (l);
    increase_precision (f, l); }
  virtual C next () { return Op::op (f[this->n], x); }
};

template<typename Op,typename C,typename V,typename X> inline Series
extrary_series (const Series& f, const X& x) {
  return (Series_rep*) new extrary_series_rep<Op,C,V,X> (f, x);
}

template<typename Op,typename C,typename V,typename X>
class extrary_implicit_series_rep : public Implicit_series_rep {
protected:
  Series f;
  X x;
public:
  extrary_implicit_series_rep (const Series& f2, const X& x2):
    f (f2), x (x2) {}
  generic expression (const generic& z) const {
    return Op::op (flatten (f, z), flatten (x)); }
  virtual void Increase_precision (nat l) {
    Implicit_series_rep::Increase_precision (l);
    increase_precision (f, l); }
  Series initialize () {
    this->initial (0)= Op::op (f[0], x);
    return Op::def (Series (this->me ()), f, x);
  }
};

template<typename Op,typename C,typename V,typename X> inline Series
extrary_implicit_series (const Series& f, const X& x) {
  Series_rep* rep= new extrary_implicit_series_rep<Op,C,V,X> (f, x);
  return implicit (Series (rep));
}

/******************************************************************************
* Coefficient-wise operations
******************************************************************************/

TMPL Series
operator + (const Series& f, const Series& g) {
  return binary_series<add_op> (f, g);
}

TMPL Series
operator - (const Series& f) {
  return unary_series<neg_op> (f);
}

TMPL Series
operator - (const Series& f, const Series& g) {
  return binary_series<sub_op> (f, g);
}

TMPL Series
operator * (const Series& f, const C& c) {
  return extrary_series<sc_mul_op<false> > (f, c);
}

TMPL Series
operator * (const C& c, const Series& f) {
  return extrary_series<sc_mul_op<true> > (f, c);
}

TMPL Series
operator / (const Series& f, const C& c) {
  return extrary_series<sc_div_op<false> > (f, c);
}

TMPL Series
quo (const Series& f, const C& c) {
  return extrary_series<sc_quo_op<false> > (f, c);
}

TMPL Series
rem (const Series& f, const C& c) {
  return extrary_series<sc_rem_op<false> > (f, c);
}

/******************************************************************************
* Other linear time operations
******************************************************************************/

TMPL
class shift_exp_series_rep : public Series_rep {
  Series f;
  int shift;
public:
  shift_exp_series_rep (const Series& f2, int shift2):
    f (f2), shift (shift2) {}
  C next () { return shift > ((int) this->n)? 0: f[this->n - shift]; }
  generic expression (const generic& z) const {
    return shift_exp (flatten (f, z), flatten (shift)); }
  void Increase_precision (nat l) {
    Series_rep::Increase_precision (l);
    increase_precision (f, max (0, ((int) l) - shift)); }
};

TMPL Series
shift_exp (const Series& f, int shift) {
  return (Series_rep*) new shift_exp_series_rep<C,V> (f, shift);
}

TMPL
class derive_series_rep : public Series_rep {
  Series f;
public:
  derive_series_rep (const Series& f2): f (f2) {}
  C next () { return C (this->n + 1) * f[this->n + 1]; }
  generic expression (const generic& z) const {
    return derive (flatten (f, z)); }
  void Increase_precision (nat l) {
    Series_rep::Increase_precision (l);
    increase_precision (f, l+1); }
};

TMPL Series
derive (const Series& f) {
  return (Series_rep*) new derive_series_rep<C,V> (f);
}

TMPL Series
xderive (const Series& f) {
  return shift_exp (derive (f), 1);
}

template<typename C, typename V, typename T> Series
derive_coeffs (const Series& f, const T& v) {
  return extrary_series<derive_op> (f, v);
}

TMPL
class integrate_series_rep : public Series_rep {
  Series f;
  C c;
public:
  integrate_series_rep (const Series& f2, const C& c2= 0): f (f2), c (c2) {}
  C next () { return this->n == 0? c: f[this->n - 1] / this->n; }
  generic expression (const generic& z) const {
    if (c == C(0)) return integrate (flatten (f, z));
    else return integrate_init (flatten (f, z), flatten (c)); }
  void Increase_precision (nat l) {
    Series_rep::Increase_precision (l);
    increase_precision (f, max (0, ((int) l) - 1)); }
};

TMPL Series
integrate (const Series& f) {
  return (Series_rep*) new integrate_series_rep<C,V> (f);
}

TMPL Series
integrate_init (const Series& f, const C& c) {
  return (Series_rep*) new integrate_series_rep<C,V> (f, c);
}

template<typename C, typename V, typename T> Series
integrate_coeffs (const Series& f, const T& v) {
  return extrary_series<integrate_op> (f, v);
}

TMPL
class dilate_series_rep : public Series_rep {
  Series f;
  nat p;
public:
  dilate_series_rep (const Series& f2, int p2): f (f2), p (p2) {}
  C next () { return this->n % p == 0? f[this->n / p]: 0; }
  generic expression (const generic& z) const {
    return dilate (flatten (f, z), flatten (p)); }
  void Increase_precision (nat l) {
    Series_rep::Increase_precision (l);
    increase_precision (f, l/p); }
};

TMPL Series
dilate (const Series& f, nat p) {
  return (Series_rep*) new dilate_series_rep<C,V> (f, p);
}

/******************************************************************************
* Multiplication and division
******************************************************************************/

TMPL
class mul_series_rep: public binary_series_rep<mul_op,C,V> {
public:
  inline mul_series_rep (const Series& f, const Series& g):
    binary_series_rep<mul_op,C,V> (f, g) {}
  C next () {
    nat i;
    C acc= this->f[0] * this->g[this->n];
    for (i=1; i<=this->n; i++)
      acc += this->f[i] * this->g[this->n-i];
    return acc;
  }
};

TMPL Series
operator * (const Series& f, const Series& g) {
  return (Series_rep*) new mul_series_rep<C,V> (f, g);
}

TMPL Series
operator / (const Series& f, const Series& g) {
  return binary_implicit_series<div_op> (f, g);
}

TMPL Series
quo (const Series& f, const Series& g) {
  return binary_implicit_series<quo_op> (f, g);
}

TMPL Series
rem (const Series& f, const Series& g) {
  return f - g * quo (f, g);
}

ARITH_INT_SUGAR(TMPL,Series,C);

/******************************************************************************
* Elementary functions
******************************************************************************/

TMPL Series
sqrt (const Series& f) {
  return unary_implicit_series<sqrt_op> (f);
}

TMPL Series
sqrt_init (const Series& f, const C& c) {
  return unary_implicit_series<sqrt_op> (f, c);
}

TMPL Series
exp (const Series& f) {
  return unary_implicit_series<exp_op> (f);
}

TMPL Series
log (const Series& f) {
  return unary_implicit_series<log_op> (f);
}

TMPL Series
log_init (const Series& f, const C& c) {
  return unary_implicit_series<log_op> (f, c);
}

TMPL Series
pow (const Series& f, const C& c) {
  return exp (c * log (f));
}

// NOTE: cos, sin and tan defined in linear_series.hpp

TMPL Series
acos (const Series& f) {
  return unary_implicit_series<acos_op> (f);
}

TMPL Series
acos_init (const Series& f, const C& c) {
  return unary_implicit_series<acos_op> (f, c);
}

TMPL Series
asin (const Series& f) {
  return unary_implicit_series<asin_op> (f);
}

TMPL Series
asin_init (const Series& f, const C& c) {
  return unary_implicit_series<asin_op> (f, c);
}

TMPL Series
atan (const Series& f) {
  return unary_implicit_series<atan_op> (f);
}

TMPL Series
atan_init (const Series& f, const C& c) {
  return unary_implicit_series<atan_op> (f, c);
}

#undef TMPL_DEF
#undef TMPL
#undef Series
#undef Series_rep
#undef Implicit_series_rep
} // namespace mmx
#endif // __MMX_SERIES_HPP
