
/******************************************************************************
* MODULE     : dense_vector.hpp
* DESCRIPTION: Dense vectors
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_DENSE_VECTOR_HPP
#define __MMX_DENSE_VECTOR_HPP
#include "vector.hpp"
#include "segment.hpp"
#include "exception.hpp"
namespace mmx {
#define TMPL template<typename C, typename V>
#define Variant dense_vector<V>
#define Vector vector<C,Variant >
#define Vector_rep vector_rep<C,Variant >
#define Abs_vector vector<Abs_type(C),Variant >
TMPL inline nat dim (const Vector& v);
TMPL inline C* seg (Vector& v);
TMPL inline const C* seg (const Vector& v);
TMPL inline bool is_scalar    (const Vector& v);
TMPL inline bool is_vectorial (const Vector& v);

/******************************************************************************
* Dense vector variants
******************************************************************************/

template<typename V>
class dense_vector {
public:
  static const nat def_len = 0;
  static const nat init_len= 1;
};

template<typename V, typename S> class linear {};
template<typename V, typename S>
class dense_vector<linear<V,S> > {
public:
  static const nat def_len = S::val;
  static const nat init_len= S::val;
};

/******************************************************************************
* Dense vectors of arbitrary sizes
******************************************************************************/

template<typename C, typename V>
class Vector_rep REP_STRUCT {
  C*   a;           // entries of vector
  nat  n;           // dimension of vector
  nat  l;           // allocated number of entries
  bool scalar_flag; // is the vector a scalar value?
public:
  inline vector_rep (C* a2, nat n2, bool flag):
    a (a2), n (n2), l (n2), scalar_flag (flag) {}
  inline ~vector_rep () { delete[] a; }
  void resize (nat n2) {
    if (n2 > l) {
      nat l2= max (n2, l<<1);
      C* b= new C[l2];
      segment_copy (b, a, n);
      delete[] a;
      a= b;
      l= l2;
    }
    n= n2;
  }
  friend class Vector;
  friend nat dim LESSGTR (const Vector& v);
  friend C* seg LESSGTR (Vector& v);
  friend const C* seg LESSGTR (const Vector& v);
  friend bool is_scalar    LESSGTR (const Vector& v);
  friend bool is_vectorial LESSGTR (const Vector& v);
};

template<typename C, typename V>
class Vector {
INDIRECT_PROTO_2 (vector, vector_rep, C, Variant)
public:
  vector () {
    nat n= Variant::def_len;
    rep= new Vector_rep (new C[n], n, false);
  }
  template<typename T> vector (const T& c) {
    nat n= Variant::init_len;
    rep= new Vector_rep (new C[n], n, true);
    segment_set (rep->a, C(c), n);
  }
  template<typename T, typename W> vector (const vector<T,W>& v) {
    C* s= new C [dim (v)]; 
    rep= new Vector_rep (s, dim (v), is_scalar (v));
    segment_cast (rep->a, seg (v), dim (v));
  }
  vector (const C& x) {
    nat n= Variant::init_len;
    rep= new Vector_rep (new C[n], n, true);
    segment_set (rep->a, x, n);
  }
  vector (const C& x, nat n) {
    rep= new Vector_rep (new C[n], n, false);
    segment_set (rep->a, x, n);
  }
  inline vector (C* a, nat n=Variant::def_len) {
    rep= new Vector_rep (a, n, false);
  }
  vector (const iterator<C>& it) {
    nat i, n= Variant::def_len;
    if (n==0) n=1;
    C* a= new C[n];
    for (i=0; busy (it); i++, ++it) {
      if (i >= n) {
	C* b= new C[n<<1];
	segment_copy (b, a, n);
	delete[] a;
	a= b;
	n= n<<1;
      }
      a[i]= *it;
    }
    rep= new Vector_rep (a, i, false);
  }

  inline C operator [] (nat i) const {
    if (is_scalar (*this)) return rep->a[0];
    return rep->a[i]; }
  inline C& operator () (nat i) {
    assert (is_vectorial (*this));
    secure(); return rep->a[i]; }
  inline C& scalar_ref () {
    assert (is_scalar (*this));
    return rep->a[0]; }

  Vector& operator << (const C& append);
  Vector& operator << (const Vector& w);
};
INDIRECT_IMPL_2 (vector, vector_rep, typename C, C, typename V, Variant)

TMPL inline nat dim (const Vector& v) { return v->n; }
TMPL inline bool is_scalar (const Vector& v) { return v->scalar_flag; }
TMPL inline bool is_vectorial (const Vector& v) { return !v->scalar_flag; }
TMPL inline const C* seg (const Vector& v) { return v->a; }
TMPL inline C* seg (Vector& v) {
  assert (is_vectorial (v));
  v.secure(); return v->a; }
TMPL inline Vector extend (const Vector& v, const Vector& w) {
  assert (is_scalar (v) && is_vectorial (w));
  return Vector (v[0], dim (w)); }

STYPE_TO_TYPE(TMPL,scalar_type,Vector,C);
STYPE_TO_TYPE(TMPL,abs_type,Vector,Abs_vector);

/******************************************************************************
* Fixed size dense vectors
******************************************************************************/

#define LVector vector<C,dense_vector<linear<V,S> > >
#define LVector_rep vector_rep<C,dense_vector<linear<V,S> > >

template<typename C, typename V, typename S>
class LVector_rep REP_STRUCT {
  C* a;
public:
  inline vector_rep (C* a2, nat dummy1, bool dummy2): a (a2) {
    (void) dummy1; (void) dummy2; }
  inline virtual ~vector_rep () { delete[] a; }
  friend class LVector;
  friend nat dim LESSGTR (const LVector& v);
  friend C*  seg LESSGTR (LVector& v);
  friend const C* seg LESSGTR (const LVector& v);
  friend bool is_scalar    LESSGTR (const Vector& v);
  friend bool is_vectorial LESSGTR (const Vector& v);
};

template<typename C, typename V, typename S> inline nat
dim (const LVector& v) { return S::val; }
template<typename C, typename V, typename S> inline bool
is_scalar (const LVector& v) { (void) v; return false; }
template<typename C, typename V, typename S> inline bool
is_vectorial (const LVector& v) { (void) v; return true; }

#undef LVector
#undef LVector_rep

/******************************************************************************
* Explicit dense vectors
******************************************************************************/

template<typename C> vector<C>
vec () {
  C* a= new C[0];
  return vector<C> (a, 0);
}

template<typename C> vector<C>
vec (const C& c1) {
  C* a= new C[1];
  a[0]= c1;
  return vector<C> (a, 1);
}

template<typename C> vector<C>
vec (const C& c1, const C& c2) {
  C* a= new C[2];
  a[0]= c1; a[1]= c2;
  return vector<C> (a, 2);
}

template<typename C> vector<C>
vec (const C& c1, const C& c2, const C& c3) {
  C* a= new C[3];
  a[0]= c1; a[1]= c2; a[2]= c3;
  return vector<C> (a, 3);
}

template<typename C> vector<C>
vec (const C& c1, const C& c2, const C& c3, const C& c4) {
  C* a= new C[4];
  a[0]= c1; a[1]= c2; a[2]= c3; a[3]= c4;
  return vector<C> (a, 4);
}

template<typename C> vector<C>
vec (const C& c1, const C& c2, const C& c3, const C& c4, const C& c5) {
  C* a= new C[5];
  a[0]= c1; a[1]= c2; a[2]= c3; a[3]= c4; a[4]= c5;
  return vector<C> (a, 5);
}

template<typename C> vector<C>
vec (const C& c1, const C& c2, const C& c3,
     const C& c4, const C& c5, const C& c6)
{
  C* a= new C[6];
  a[0]= c1; a[1]= c2; a[2]= c3; a[3]= c4; a[4]= c5; a[5]= c6;
  return vector<C> (a, 6);
}

/******************************************************************************
* Extracting ranges of elements and appending new elements
******************************************************************************/

TMPL Vector
range (const Vector& v, nat start, nat end) {
  if (is_scalar (v)) return Vector (v[0], end-start);
  nat n= end- start;
  C* a= new C[n];
  segment_copy (a, seg (v) + start, n);
  return Vector (a, n);
}

TMPL inline C car (const Vector& v) { return v[0]; }
TMPL inline Vector cdr (const Vector& v) { return range (v, 1, dim (v)); }

TMPL Vector
append (const Vector& v, const Vector& w) {
  assert (is_vectorial (v) && is_vectorial (w));
  nat n= dim (v), p= dim (w);
  C* a= new C[n+p];
  segment_copy (a, seg (v), n);
  segment_copy (a+n, seg (w), p);
  return Vector (a, n+p);
}

TMPL Vector
cons (const C& c, const Vector& v) {
  assert (is_vectorial (v));
  nat n= dim (v);
  C* a= new C[n+1];
  a[0]= c;
  segment_copy (a+1, seg (v), n);
  return Vector (a, n+1);
}

TMPL Vector&
Vector::operator << (const C& append) {
  assert (is_vectorial (*this));
  secure ();
  nat n= rep->n;
  rep->resize (n+1);
  rep->a[n]= append;
  return *this;
}

TMPL Vector&
Vector::operator << (const Vector& w) {
  assert (is_vectorial (*this) && is_vectorial (w));
  secure ();
  nat n= rep->n, p= dim (w);
  rep->resize (n + p);
  segment_copy (rep->a + n, seg (w), p);
  return *this;
}

/******************************************************************************
* Routines on dense vectors
******************************************************************************/

template<typename Op,typename C,typename V> inline Vector
unary_vector (const Vector& v) {
  if (is_scalar (v)) return Vector (Op::op (v[0]));
  nat n= dim (v);
  C* r= new C[n];
  segment_unary<Op> (r, seg (v), n);
  return Vector (r, n);
}

template<typename Op,typename C,typename V> inline Vector
binary_vector (const Vector& v, const Vector& w) {
  if (is_scalar (v) || is_scalar (w)) {
    if (is_vectorial (v)) return binary_vector<Op> (v, extend (w, v));
    if (is_vectorial (w)) return binary_vector<Op> (extend (v, w), w);
    return Vector (Op::op (v[0], w[0]));
  }
  nat n= dim (v);
  ASSERT_DOMAIN (dim (w) == n);
  C* r= new C[n];
  segment_binary<Op> (r, seg (v), seg (w), n);
  return Vector (r, n);
}

template<typename Op,typename C,typename V,typename X> inline Vector
extrary_vector (const Vector& v, const X& x) {
  if (is_scalar (v)) return Vector (Op::op (v[0], x));
  nat n= dim (v);
  C* r= new C[n];
  segment_extrary<Op> (r, seg (v), x, n);
  return Vector (r, n);
}

template<typename Op,typename C,typename V> inline Vector&
in_place_binary_vector (Vector& v, const Vector& w) {
  if (is_scalar (v) || is_scalar (w)) {
    if (is_vectorial (v))
      return in_place_binary_vector<Op> (v, extend (w, v));
    if (is_vectorial (w))
      v= extend (v, w);
    else {
      Op::in_place (v.scalar_ref (), w[0]);
      return v;
    }
  }
  nat n= dim (v);
  ASSERT_DOMAIN (dim (w) == n);
  segment_in_place_binary<Op> (seg (v), seg (w), n);
  return v;
}

template<typename Op,typename C,typename V,typename X> inline Vector&
in_place_extrary_vector (Vector& v, const X& x) {
  if (is_scalar (v)) {
    Op::in_place (v.scalar_ref (), x);
    return v;
  }
  nat n= dim (v);
  segment_in_place_extrary<Op> (seg (v), x, n);
  return v;
}

template<typename Op,typename C,typename V> inline bool
test_vector (const Vector& v, const Vector& w) {
  if (is_scalar (v) || is_scalar (w)) {
    if (is_vectorial (v)) return test_vector<Op> (v, extend (w, v));
    if (is_vectorial (w)) return test_vector<Op> (extend (v, w), w);
    return Op::op (v[0], w[0]);
  }
  nat n= dim (v);
  if (dim (w) != n) return false;
  return segment_test<Op> (seg (v), seg (w), n);
}

template<typename Op,typename C,typename V,typename X> inline bool
test_extra_vector (const Vector& v, const X& c) {
  if (is_scalar (v)) return Op::op (v[0], c);
  return segment_test_extra<Op> (seg (v), c, dim (v));
}

template<typename Op, typename C, typename V> C
big (const Vector& v) {
  if (is_scalar (v)) {
    assert (v[0] == Op::template neutral<C> ());
    return v[0];
  }
  return segment_big<Op> (seg (v), dim (v));
}

/******************************************************************************
* Instantiations of the abstract routines
******************************************************************************/

TMPL Vector copy (const Vector& v) {
  return unary_vector<copy_op> (v); }
TMPL Vector melt (const Vector& v) {
  return unary_vector<melt_op> (v); }
TMPL Vector fuse (const Vector& v) {
  return unary_vector<fuse_op> (v); }
TMPL Vector operator - (const Vector& v) {
  return unary_vector<neg_op> (v); }
TMPL Vector operator + (const Vector& v, const Vector& w) {
  return binary_vector<add_op> (v, w); }
TMPL Vector operator - (const Vector& v, const Vector& w) {
  return binary_vector<sub_op> (v, w); }
TMPL Vector operator * (const Vector& v, const Vector& w) {
  return binary_vector<mul_op> (v, w); }
TMPL Vector operator / (const Vector& v, const Vector& w) {
  return binary_vector<div_op> (v, w); }
TMPL Vector quo (const Vector& v, const Vector& w) {
  return binary_vector<quo_op> (v, w); }
TMPL Vector rem (const Vector& v, const Vector& w) {
  return binary_vector<rem_op> (v, w); }
TMPL Vector operator * (const Vector& v, const C& c) {
  return extrary_vector<sc_mul_op<false> > (v, c); }
TMPL Vector operator * (const C& c, const Vector& v) {
  return extrary_vector<sc_mul_op<true> > (v, c); }
TMPL Vector operator / (const Vector& v, const C& c) {
  return extrary_vector<sc_div_op<false> > (v, c); }
TMPL Vector operator / (const C& c, const Vector& v) {
  return extrary_vector<sc_div_op<true> > (v, c); }
TMPL Vector quo (const Vector& v, const C& c) {
  return extrary_vector<sc_quo_op<false> > (v, c); }
TMPL Vector rem (const Vector& v, const C& c) {
  return extrary_vector<sc_rem_op<false> > (v, c); }
TMPL Vector& operator += (Vector& v, const Vector& w) {
  return in_place_binary_vector<add_op> (v, w); }
TMPL Vector& operator -= (Vector& v, const Vector& w) {
  return in_place_binary_vector<sub_op> (v, w); }
TMPL Vector& operator *= (Vector& v, const Vector& w) {
  return in_place_binary_vector<mul_op> (v, w); }
TMPL Vector& operator /= (Vector& v, const Vector& w) {
  return in_place_binary_vector<div_op> (v, w); }
template<typename C,typename V,typename K> Vector&
operator *= (Vector& v, const K& c) {
  return in_place_extrary_vector<sc_mul_op<false> > (v, c); }
template<typename C,typename V,typename K> Vector&
operator /= (Vector& v, const K& c) {
  return in_place_extrary_vector<sc_div_op<false> > (v, c); }
TMPL bool eq (const Vector& v, const Vector& w) {
  return test_vector<eq_op> (v, w); }
TMPL inline bool neq (const Vector& v, const Vector& w) {
  return !eq (v, w); }
TMPL bool operator == (const Vector& v, const Vector& w) {
  return test_vector<equal_op> (v, w); }
TMPL bool operator <= (const Vector& v, const Vector& w) {
  return test_vector<lesseq_op> (v, w); }
TMPL bool operator >= (const Vector& v, const Vector& w) {
  return test_vector<gtreq_op> (v, w); }
template<typename C,typename V,typename K> bool
operator == (const Vector& v, const K& c) {
  return test_extra_vector<equal_op> (v, c); }
template<typename C,typename V,typename K> bool
operator <= (const Vector& v, const K& c) {
  return test_extra_vector<lesseq_op> (v, c); }
template<typename C,typename V,typename K> bool
operator >= (const Vector& v, const K& c) {
  return test_extra_vector<gtreq_op> (v, c); }

ARITH_INT_SUGAR(TMPL,Vector,C);
STRICT_COMPARE_SUGAR(TMPL,Vector);
COMPARE_INT_SUGAR(TMPL,Vector,Vector);

TMPL Vector derive (const Vector& v) {
  return unary_vector<derive_op> (v); }
TMPL Vector integrate (const Vector& v) {
  return unary_vector<integrate_op> (v); }
template<typename C,typename V,typename X> Vector
derive (const Vector& v, const X& x) {
  return extrary_vector<derive_op> (v, x); }
template<typename C,typename V,typename X> Vector
integrate (const Vector& v, const X& x) {
  return extrary_vector<integrate_op> (v, x); }

TMPL inline C big_add (const Vector& v) { return big<add_op> (v); }
TMPL inline C big_mul (const Vector& v) { return big<mul_op> (v); }
TMPL inline C big_or  (const Vector& v) { return big< or_op> (v); }
TMPL inline C big_and (const Vector& v) { return big<and_op> (v); }
TMPL inline C big_min (const Vector& v) { return big<min_op> (v); }
TMPL inline C big_max (const Vector& v) { return big<max_op> (v); }

/******************************************************************************
* Operations with cast
******************************************************************************/

template<typename Op,typename R,typename C,typename V>
inline vector<R,Variant >
cast_unary_vector (const Vector& v) {
  if (is_scalar (v))
    return vector<R,Variant > (Op::template cast_helper<R>::cast_op (v[0]));
  nat n= dim (v);
  R* r= new R[n];
  segment_cast_unary<Op> (r, seg (v), n);
  return vector<R,Variant > (r, n);
}

template<typename Op,typename R,typename S,typename V>
struct cast_unary_helper<Op,vector<R,Variant >,vector<S,Variant > > {
  static inline vector<R,Variant >
  op (const vector<S,Variant>& x) {
    return cast_unary_vector<Op,R> (x); }
};

template<typename Op,typename R,typename C,typename V,typename X>
inline vector<R,Variant >
cast_extrary_vector (const Vector& v, const X& x) {
  if (is_scalar (v))
    return vector<R,Variant > (Op::template cast_helper<R>::cast_op (v[0], x));
  nat n= dim (v);
  R* r= new R[n];
  segment_cast_extrary<Op> (r, seg (v), x, n);
  return vector<R,Variant > (r, n);
}

template<typename Op,typename R,typename S1,typename S2,typename V>
struct cast_binary_helper<Op,vector<R,Variant >,vector<S1,Variant >,S2> {
  static inline vector<R,Variant >
  op (const vector<S1,Variant>& x, const S2& y) {
    return cast_extrary_vector<typename Op::sc_right,R> (x, y); }
};

template<typename Op,typename R,typename S1,typename S2,typename V>
struct cast_binary_helper<Op,vector<R,Variant>,S1,vector<S2,Variant> > {
  static inline vector<R,Variant>
  op (const S2& x, const vector<S1,Variant>& y) {
    return cast_extrary_vector<typename Op::sc_left,R> (y, x); }
};

TMPL Abs_vector abs (const Vector& v) {
  return cast_abs<Abs_vector > (v); }

/******************************************************************************
* Implementation of some routines on operators
******************************************************************************/

template<typename C> inline C
trig_op::diff_op (const C& me, const C& x) {
  return from_vector (vec (-derive (access (x, 0)) * access (me, 1),
			   derive (access (x, 0)) * access (me, 0)));
}

template<typename C> inline C
tan_op::def (const C& me, const C& f) {
  (void) me; vector<C> v= cos_sin (f); return v[1] / v[0];
}

template<typename C> inline C
solve_vector_lde_op::diff_op (const C& me, const C& x) {
  typedef As_vector_type (C) V;
  typedef Scalar_type (V) K;
  V w= as_vector (me);
  V v= as_vector (x);
  nat i, n= dim (v);
  K sum (0);
  V d (K (0), n);
  for (i=0; i<n; i++) {
    if (i==0) sum= v[i] * w[i];
    else sum= sum + v[i] * w[i];
    if (i<n-1) d(i)= w[i+1];
    else d(i)= sum;
  }
  return from_vector (d);
}

/******************************************************************************
* Further operations
******************************************************************************/

template<typename C> vector<C> cos_sin (const C& x) {
  return vec (cos (x), sin (x)); }
template<typename C> vector<C> trig (const vector<C>& v) {
  return vec (cos (v[0]), sin (v[0])); }

TMPL C dot (const Vector& v, const Vector& w) {
  assert (is_vectorial (v) && is_vectorial (w) && dim (v) == dim (w));
  return segment_inn_prod<C> (seg (v), seg (w), dim (v)); }

#undef TMPL
#undef Variant
#undef Vector
#undef Vector_rep
#undef Abs_vector
} // namespace mmx
#endif // __MMX_DENSE_VECTOR_HPP
