
/******************************************************************************
* MODULE     : anfunc.hpp
* DESCRIPTION: Analytic functions
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_ANFUNC_HPP
#define __MMX_ANFUNC_HPP
#include "series.hpp"
//#include "relaxed_series.hpp"
#include "dense_vector.hpp"
namespace mmx {
class std_anfunc {};
#define TMPL_DEF template<typename C,typename V= std_anfunc>
#define TMPL template<typename C,typename V>
#define Anfunc anfunc<C,V>
#define Anfunc_rep anfunc_rep<C,V>
#define Implicit_anfunc_rep implicit_anfunc_rep<C,V>
#define Assume assume_bound<typename unvectorize<Error >::val>
#define Scalar typename unvectorize<C>::val
#define Series_variant naive_multiply
//#define Series_variant relaxed<naive_multiply>
#define Series series<C,Series_variant>
#define Error Abs_type(C)
TMPL class anfunc;
typedef long long unsigned int llnat;

/******************************************************************************
* The analytic function class
******************************************************************************/

TMPL
class anfunc_cache {
public:
  Series              expansion;
  table<llnat,nat>    assumption;
  table<Error,nat>    head_bound;
  table<Error,nat>    tail_bound;
  table<Error,nat>    conv_bound;
  table<Anfunc,void*> focus_at;
  table<Anfunc,void*> move_to;
};

template<typename C>
class assume_bound {
public:
  llnat serial;
  table<vector<C>,void*> assumption;
  table<vector<C>,void*> conclusion;
  inline assume_bound () {
    static llnat current_serial= 1;
    serial= current_serial++; }
};

template<typename C>
struct unvectorize {
  typedef C val;
  static inline vector<val> encode (const C& x) {
    return vector<val> (x, 1); }
  static inline C decode (const vector<val>& v, const C& gauge) {
    (void) gauge;
    return v[0]; }
};

TMPL_DEF
class anfunc_rep REP_STRUCT {
public:
  anfunc_cache<C,V>* cache;
  void init_cache ();

public:
  inline anfunc_rep (): cache (NULL) {}
  inline virtual ~anfunc_rep () { if (cache != NULL) delete cache; }
  virtual generic expression (const generic& z) const;

public:
  virtual Series Expand () const = 0;
  virtual Error Tail_bound (nat order, Assume& a) const = 0;
  virtual Anfunc Focus (const Scalar& z) const = 0;
  virtual Anfunc Move (const Scalar& z) const = 0;
  virtual Anfunc Derive () const = 0;
  inline Anfunc me () const;
  friend class Anfunc;

  virtual C initial () const;
  virtual Anfunc equation () const;
};

TMPL_DEF
class anfunc {
INDIRECT_PROTO_2 (anfunc, anfunc_rep, C, V)
public:
  anfunc (int c= 0);
  anfunc (const C& c);
  anfunc (const polynomial<C>& P);
  inline C operator [] (nat n) const;
};
INDIRECT_IMPL_2 (anfunc, anfunc_rep, typename C, C, typename V, V)

STYPE_TO_TYPE(TMPL,scalar_type,Anfunc,C);

/******************************************************************************
* Forward definitions of operations on analytic functions
******************************************************************************/

TMPL Anfunc operator + (const Anfunc& f, const Anfunc& g);
TMPL Anfunc operator - (const Anfunc& f);
TMPL Anfunc operator - (const Anfunc& f, const Anfunc& g);
template<typename C,typename V,typename K> Anfunc operator * (const K& c, const Anfunc& f); 
TMPL Anfunc operator * (const Anfunc& f, const C& c);
TMPL Anfunc operator / (const Anfunc& f, const C& c);
TMPL Anfunc operator * (const Anfunc& f, const Anfunc& g);
TMPL Anfunc operator / (const Anfunc& f, const Anfunc& g);

TMPL Anfunc integrate (const Anfunc& f);
TMPL Anfunc integrate_init (const Anfunc& f, const C& c);

TMPL Anfunc sqrt (const Anfunc& f);
TMPL Anfunc sqrt_init (const Anfunc& f, const C& c);
TMPL Anfunc exp (const Anfunc& f);
TMPL Anfunc log (const Anfunc& f);
TMPL Anfunc log_init (const Anfunc& f, const C& c);
// cos, sin and tan are defined in linear_series.hpp
TMPL Anfunc acos (const Anfunc& f);
TMPL Anfunc asin (const Anfunc& f);
TMPL Anfunc atan (const Anfunc& f);

/******************************************************************************
* Basic routines for analytic functions
******************************************************************************/

TMPL void
Anfunc_rep::init_cache () {
  if (cache == NULL) {
    cache= new anfunc_cache<C,V> ();
    cache->expansion = Expand ();
    cache->assumption= table<llnat,nat> (0);
  }
}

TMPL Series
expand (const Anfunc& f) {
  Anfunc_rep* rep= const_cast<Anfunc_rep*> (f.operator -> ());
  rep->init_cache ();
  return rep->cache->expansion;
}

TMPL Series
expand (const Anfunc& f, nat order) {
  Series r= expand (f);
  if (order>0) (void) r[order-1];
  return r;
}

TMPL Error
tail_bound (const Anfunc& f, nat order, Assume& a) {
  Anfunc_rep* rep= const_cast<Anfunc_rep*> (f.operator -> ());
  rep->init_cache ();
  if (rep->cache->assumption[order] != a.serial) {
    (void) expand (f, order);
    Error r= f->Tail_bound (order, a);
    rep->cache->assumption (order)= a.serial;
    rep->cache->tail_bound (order)= r;
  }
  return rep->cache->tail_bound [order];
}

template<typename C,typename T> C
minimum (const table<vector<C>,T>& t) {
  iterator<T> it= entries (t);
  C m= t[*it][0];
  for (++it; busy (it); ++it)
    for (nat i=0; i<dim(t[*it]); i++)
      m= min (m, t[*it][i]);
  return m;
}

TMPL Error
tail_bound (const Anfunc& f, nat order) {
  Assume a1;
  Error r= tail_bound (f, order, a1);
  if (N (a1.assumption) == 0) return r;
  nat iterations= (order >> 2) + 4;
  for (nat i=0; i<2*iterations; i++) {
    //cout << "Pass " << i << "\n";
    // One step with a simple iteration
    Assume a2;
    a2.assumption= a1.conclusion;
    r= tail_bound (f, order, a2);
    //cout << a2.assumption << " -> " << a2.conclusion << "\n";
    if (!(a2.assumption <= a2.conclusion)) return Infinity (Error);
    if (a2.assumption == a2.conclusion) return r;

    // One secant-method iteration, if possible
    Assume a3;
    table<vector<Error >,void*> d1= a1.conclusion - a1.assumption;
    table<vector<Error >,void*> d2= a2.conclusion - a2.assumption;
    if (d1 > d2) {
      Error l = minimum (d1 / (d1 - d2));
      a3.assumption= a1.assumption + l * (a1.conclusion - a1.assumption);
      r= tail_bound (f, order, a3);
      //cout << a3.assumption << " -> " << a3.conclusion << "\n";
      if (!(a3.assumption <= a3.conclusion)) return Infinity (Error);
      if (a3.assumption == a3.conclusion) return r;
    }
    else a3= a2;

    // Check whether we are done
    if (i >= iterations) {
      Assume a4;
      //a4.assumption= a3.conclusion + (a3.conclusion - a3.assumption);
      a4.assumption= Error (1.001) * a3.conclusion;
      r= tail_bound (f, order, a4);
      //cout << a4.assumption << " -> " << a4.conclusion << "\n";
      if (a4.conclusion <= a4.assumption) return r;
    }

    // Iterate
    a1.assumption= a3.assumption;
    a1.conclusion= a3.conclusion;
  }
  return Infinity (Error);
}

TMPL Error
head_bound (const Anfunc& f, nat order) {
  Anfunc_rep* rep= const_cast<Anfunc_rep*> (f.operator -> ());
  rep->init_cache ();
  if (!rep->cache->head_bound->contains (order)) {
    Series s  = expand (f, order);
    Error  sum= 0;
    for (nat i=0; i<order; i++)
      sum += abs (s[i]);
    rep->cache->head_bound (order)= sum;
  }
  return rep->cache->head_bound [order];
}

TMPL Error
bound (const Anfunc& f, nat order, Assume& a) {
  return head_bound (f, order) + tail_bound (f, order, a);
}

TMPL Error
conv_bound (const Anfunc& me, const Anfunc& f, const Anfunc& g,
	    nat order)
{
  Anfunc_rep* rep= const_cast<Anfunc_rep*> (f.operator -> ());
  rep->init_cache ();
  if (!rep->cache->conv_bound->contains (order)) {
    Error sum=0, cumul= 0;
    Series ff= expand (f, order);
    Series gg= expand (g, order);
    for (nat i=1; i<order; i++) {
      cumul += abs (gg[order-i]);
      sum   += abs (ff[i]) * cumul;
    }
    rep->cache->conv_bound (order)= sum;
  }
  return rep->cache->conv_bound [order];
}

TMPL Anfunc
derive (const Anfunc& f) {
  return f->Derive ();
}

TMPL Anfunc
focus (const Anfunc& f, const Scalar& z) {
  Anfunc_rep* rep= const_cast<Anfunc_rep*> (f.operator -> ());
  rep->init_cache ();
  if (!rep->cache->focus_at->contains ((void*) &z))
    rep->cache->focus_at ((void*) &z)= rep->Focus (z);
  return rep->cache->focus_at [(void*) &z];
}

TMPL Anfunc
move (const Anfunc& f, const Scalar& z) {
  Anfunc_rep* rep= const_cast<Anfunc_rep*> (f.operator -> ());
  rep->init_cache ();
  if (!rep->cache->move_to->contains ((void*) &z))
    rep->cache->move_to ((void*) &z)= rep->Move (z);
  return rep->cache->move_to [(void*) &z];
}

TMPL C
eval (const Anfunc& f, const Scalar& z) {
  nat order= get_precision ();
  // FIXME: determine 'order' in a better way
  C sum= f[order];
  for (nat i=1; i<=order; i++)
    sum= sum * z + f[order-i];
  return sum;
}

/******************************************************************************
* Standard functions
******************************************************************************/

TMPL generic
var (const Anfunc& f) {
  return var (expand (f));
}

TMPL inline generic
Anfunc_rep::expression (const generic& z) const {
  return flatten (expand (me ()), z);
}

TMPL inline Anfunc
Anfunc_rep::me () const {
  return Anfunc (this, true);
}

TMPL C
Anfunc_rep::initial () const {
  assert (false);
}

TMPL Anfunc
Anfunc_rep::equation () const {
  assert (false);
}

TMPL inline C
Anfunc::operator [] (nat n) const {
  return expand (*this) [n];
}

TMPL nat
hash (const Anfunc& f) {
  Series s= expand (f);
  register nat i, h= 7531;
  for (i=0; i<4; i++)
    h= (h<<1) ^ (h<<5) ^ (h>>27) ^ hash (C (s[i]));
  return h;
}

TMPL generic
flatten (const Anfunc& f, const generic& z) {
  if (get_series_formula_output ()) return f->expression (z);
  else return flatten (expand (f), z);
}

TMPL generic
flatten (const Anfunc& f) {
  return flatten (f, var (f));
}

PRINT_SUGAR(TMPL,Anfunc)

/******************************************************************************
* Implicit analytic functions
******************************************************************************/

template<typename C,typename V>
class anfunc_implicit_series_rep:
  public implicit_series_rep<C,Series_variant>
{
protected:
  Anfunc f;
public:
  anfunc_implicit_series_rep (const Anfunc& f2): f(f2) {}
  generic expression (const generic& z) const {
    return flatten (f, z); }
  Series initialize () {
    this->initial (0)= f->initial ();
    return expand (f->equation ()); }
};

TMPL
class implicit_anfunc_rep: public Anfunc_rep {
protected:
  C      init;
  Anfunc eq;
public:
  inline implicit_anfunc_rep () {}
  virtual generic expression (const generic& z) const {
    (void) z; return "implicit"; }
  inline void initialize () {
    init= this->Initial ();
    eq  = this->Equation (); }
  C initial () const { return init; }
  Anfunc equation () const { return eq; }
  virtual C Initial () const = 0;
  virtual Anfunc Equation () const = 0;
  virtual Series Expand () const {
    series_rep<C,Series_variant>* rep=
      new anfunc_implicit_series_rep<C,V> (this->me ());
    Series s (rep);
    this->cache->expansion= s;
    Series r= implicit (s);
    this->cache->expansion= r; // redudant, but for security
    return r; }
  virtual Error Tail_bound (nat order, Assume& a) const {
    void* code= (void*) this;
    Error zero= 0 * abs (init); // gets right dimensions in vector case
    if (!a.assumption->contains (code))
      a.assumption (code)= unvectorize<Error>::encode (zero);
    this->cache->assumption (order)= a.serial;
    this->cache->tail_bound (order)=
      unvectorize<Error>::decode (a.assumption [code], zero);
    Error bnd= tail_bound (this->equation (), order, a);
    if (order == 0) bnd += abs (this->initial ());
    this->cache->tail_bound (order)= bnd; // redundant, but for security
    a.conclusion (code)= unvectorize<Error>::encode (bnd);
    return bnd; }
};

TMPL Anfunc
implicit (implicit_anfunc_rep<C,V>* rep) {
  rep->initialize ();
  return (Anfunc_rep*) rep;
}

/******************************************************************************
* Explicit anfunc functions
******************************************************************************/

TMPL
class polynomial_anfunc_rep: public Anfunc_rep {
  polynomial<C> P;
public:
  polynomial_anfunc_rep (const polynomial<C>& P2): P (P2) {}
  Series Expand () const { return Series (P); }
  Error Tail_bound (nat order, Assume& a) const {
    (void) a;
    Error  sum= 0;
    Series s  = expand (this->me (), order);
    for (nat i= order; i<dim (P); i++)
      sum += abs (s[i]);
    return sum; }
  Anfunc Focus (const Scalar& z) const {
    return Anfunc (q_difference (P, z)); }
  Anfunc Move (const Scalar& z) const {
    polynomial<Scalar> Q= seq (z, Scalar (1));
    return Anfunc (compose (P, Q)); }
  Anfunc Derive () const {
    return Anfunc (derive (P)); }
};

TMPL
Anfunc::anfunc (int c) {
  rep= new polynomial_anfunc_rep<C,V> (polynomial<C> (C (c), 0));
}

TMPL
Anfunc::anfunc (const C& c) {
  rep= new polynomial_anfunc_rep<C,V> (polynomial<C> (c, 0));
}

TMPL
Anfunc::anfunc (const polynomial<C>& P) {
  rep= new polynomial_anfunc_rep<C,V> (P);
}

/******************************************************************************
* Abstract operations on analytic functions
******************************************************************************/

template<typename Op,typename C,typename V>
class unary_anfunc_rep : public Anfunc_rep {
protected:
  Anfunc f;
public:
  inline unary_anfunc_rep (const Anfunc& f2): f(f2) {}
  Series Expand () const {
    return Op::op (expand (f)); }
  Anfunc Focus (const Scalar& z) const {
    return Op::op (focus (f, z)); }
  Anfunc Move (const Scalar& z) const {
    return Op::op (move (f, z)); }
  Anfunc Derive () const {
    return Op::diff_op (this->me (), f); }
};

template<typename Op,typename C,typename V> inline Anfunc
unary_anfunc (const Anfunc& f) {
  return (Anfunc_rep*) new unary_anfunc_rep<Op,C,V> (f);
}

template<typename Op,typename C,typename V>
class unary_implicit_anfunc_rep : public Implicit_anfunc_rep {
protected:
  Anfunc f;
  bool with_init;
  C c;
public:
  inline unary_implicit_anfunc_rep (const Anfunc& f2):
    f (f2), with_init (false) {}
  inline unary_implicit_anfunc_rep (const Anfunc& f2, const C& c2):
    f (f2), with_init (true), c (c2) {}
  generic expression (const generic& z) const {
    if (with_init) return Op::op_init (flatten (f, z), flatten (c));
    else return Op::op (flatten (f, z)); }
  C Initial () const {
    if (with_init) return c;
    else return Op::op (expand (f) [0]); }
  Anfunc Equation () const {
    return Op::def (this->me (), f); }
  Series Expand () const {
    if (with_init) return Op::op_init (expand (f), c);
    else return Op::op (expand (f)); }
  Anfunc Focus (const Scalar& z) const {
    if (with_init) return Op::op_init (focus (f, z), c);
    else return Op::op (focus (f, z)); }
  Anfunc Move (const Scalar& z) const {
    //return Op::op_init (move (f, z), eval (this->me (), z));
    if (with_init) return Op::op_init (move (f, z), c);
    else return Op::op (move (f, z)); }
  Anfunc Derive () const {
    return Op::diff_op (this->me (), f); }
};

template<typename Op,typename C,typename V> inline Anfunc
unary_implicit_anfunc (const Anfunc& f) {
  return implicit (new unary_implicit_anfunc_rep<Op,C,V> (f));
}

template<typename Op,typename C,typename V> inline Anfunc
unary_implicit_anfunc (const Anfunc& f, const C& c) {
  return implicit (new unary_implicit_anfunc_rep<Op,C,V> (f, c));
}

template<typename Op,typename C,typename V>
class binary_anfunc_rep : public Anfunc_rep {
protected:
  Anfunc f, g;
public:
  inline binary_anfunc_rep (const Anfunc& f2, const Anfunc& g2):
    f(f2), g (g2) {}
  Series Expand () const {
    return Op::op (expand (f), expand (g)); }
  Anfunc Focus (const Scalar& z) const {
    return Op::op (focus (f, z), focus (g, z)); }
  Anfunc Move (const Scalar& z) const {
    return Op::op (move (f, z), move (g, z)); }
  Anfunc Derive () const {
    return Op::diff_op (this->me (), f, g); }
};

template<typename Op,typename C,typename V> inline Anfunc
binary_anfunc (const Anfunc& f, const Anfunc& g) {
  return (Anfunc_rep*) new binary_anfunc_rep<Op,C,V> (f, g);
}

template<typename Op,typename C,typename V>
class binary_implicit_anfunc_rep : public Implicit_anfunc_rep {
protected:
  Anfunc f, g;
public:
  inline binary_implicit_anfunc_rep (const Anfunc& f2, const Anfunc& g2):
    f (f2), g (g2) {}
  generic expression (const generic& z) const {
    return Op::op (flatten (f, z), flatten (g, z)); }
  C Initial () const {
    return Op::op (expand (f), expand (g)) [0]; }
  Anfunc Equation () const {
    return Op::def (this->me (), f, g); }
  Series Expand () const {
    return Op::op (expand (f), expand (g)); }
  Anfunc Focus (const Scalar& z) const {
    return Op::op (focus (f, z), focus (g, z)); }
  Anfunc Move (const Scalar& z) const {
    return Op::op (move (f, z), move (g, z)); }
  Anfunc Derive () const {
    return Op::diff_op (this->me (), f, g); }
};

template<typename Op,typename C,typename V> inline Anfunc
binary_implicit_anfunc (const Anfunc& f, const Anfunc& g) {
  return implicit (new binary_implicit_anfunc_rep<Op,C,V> (f, g));
}

template<typename Op,typename C,typename V,typename X>
class extrary_anfunc_rep : public Anfunc_rep {
protected:
  Anfunc f;
  X x;
public:
  inline extrary_anfunc_rep (const Anfunc& f2, const X& x2):
    f(f2), x (x2) {}
  Series Expand () const {
    return Op::op (expand (f), x); }
  Anfunc Focus (const Scalar& z) const {
    return Op::op (focus (f, z), x); }
  Anfunc Move (const Scalar& z) const {
    return Op::op (move (f, z), x); }
  Anfunc Derive () const {
    return Op::diff_op (this->me (), f, x); }
};

template<typename Op,typename C,typename V,typename X> inline Anfunc
extrary_anfunc (const Anfunc& f, const X& x) {
  return (Anfunc_rep*) new extrary_anfunc_rep<Op,C,V,X> (f, x);
}

/******************************************************************************
* Coefficient-wise operations
******************************************************************************/

TMPL
class add_anfunc_rep: public binary_anfunc_rep<add_op,C,V> {
public:
  add_anfunc_rep (const Anfunc& f, const Anfunc& g):
    binary_anfunc_rep<add_op,C,V> (f, g) {}
  Error Tail_bound (nat order, Assume& a) const {
    return tail_bound (this->f, order, a) + tail_bound (this->g, order, a); }
};

TMPL Anfunc
operator + (const Anfunc& f, const Anfunc& g) {
  return (Anfunc_rep*) new add_anfunc_rep<C,V> (f, g);
}

TMPL
class neg_anfunc_rep: public unary_anfunc_rep<neg_op,C,V> {
public:
  neg_anfunc_rep (const Anfunc& f):
    unary_anfunc_rep<neg_op,C,V> (f) {}
  Error Tail_bound (nat order, Assume& a) const {
    return tail_bound (this->f, order, a); }
};

TMPL Anfunc
operator - (const Anfunc& f) {
  return (Anfunc_rep*) new neg_anfunc_rep<C,V> (f);
}

TMPL
class sub_anfunc_rep: public binary_anfunc_rep<sub_op,C,V> {
public:
  sub_anfunc_rep (const Anfunc& f, const Anfunc& g):
    binary_anfunc_rep<sub_op,C,V> (f, g) {}
  Error Tail_bound (nat order, Assume& a) const {
    return tail_bound (this->f, order, a) + tail_bound (this->g, order, a); }
};

TMPL Anfunc
operator - (const Anfunc& f, const Anfunc& g) {
  return (Anfunc_rep*) new sub_anfunc_rep<C,V> (f, g);
}

TMPL
class sc_mul_anfunc_rep: public extrary_anfunc_rep<sc_mul_op<false>,C,V,C> {
public:
  sc_mul_anfunc_rep (const Anfunc& f, const C& c):
    extrary_anfunc_rep<sc_mul_op<false>,C,V,C> (f, c) {}
  Error Tail_bound (nat order, Assume& a) const {
    return tail_bound (this->f, order, a) * abs (this->x); }
};

TMPL Anfunc
operator * (const Anfunc& f, const C& c) {
  return (Anfunc_rep*) new sc_mul_anfunc_rep<C,V> (f, c);
}

template<typename C,typename V,typename K> Anfunc
operator * (const K& c, const Anfunc& f) {
  return (Anfunc_rep*) new sc_mul_anfunc_rep<C,V> (f, c);
}

TMPL Anfunc
operator / (const Anfunc& f, const C& c) {
  return (Anfunc_rep*) new sc_mul_anfunc_rep<C,V> (f, 1/c);
}

/******************************************************************************
* Multiplication and division
******************************************************************************/

TMPL
class mul_anfunc_rep: public binary_anfunc_rep<mul_op,C,V> {
public:
  mul_anfunc_rep (const Anfunc& f, const Anfunc& g):
    binary_anfunc_rep<mul_op,C,V> (f, g) {}
  Error Tail_bound (nat order, Assume& a) const {
    return
      conv_bound (this->me (), this->f, this->g, order) +
      head_bound (this->f, order) * tail_bound (this->g, order, a) +
      tail_bound (this->f, order, a) * bound (this->g, order, a); }
};

TMPL Anfunc
operator * (const Anfunc& f, const Anfunc& g) {
  return (Anfunc_rep*) new mul_anfunc_rep<C,V> (f, g);
}

TMPL
class shift_exp_anfunc_rep: public Anfunc_rep {
protected:
  Anfunc f;
  int shift;
public:
  shift_exp_anfunc_rep (const Anfunc& f2, int shift2):
    f (f2), shift (shift2) {}
  Series Expand () const {
    return shift_exp (expand (f), shift); }
  Error Tail_bound (nat order, Assume& a) const {
    Series s= expand (this->f, order);
    Error sum= tail_bound (this->f, order, a);
    if (shift >= 0)
      for (nat i=0; i < min (order, (nat) shift); i++)
	sum += abs (s[order-1-i]);
    else
      for (nat i=0; i < ((nat) -shift); i++)
	sum += abs (s[order+i]);
    return sum; }
  Anfunc Focus (const Scalar& z) const {
    if (shift >= 0) return binpow (z, shift) * shift_exp (focus (f, z), shift);
    else return binpow (1/z, -shift) * shift_exp (focus (f, z), shift); }
  Anfunc Move (const Scalar& z) const {
    assert (false); // not yet implemented;
    return this->me (); }
  Anfunc Derive () const {
    assert (false); // not yet implemented;
    return this->me (); }
};

TMPL Anfunc
shift_exp (const Anfunc& f, int shift) {
  return (Anfunc_rep*) new shift_exp_anfunc_rep<C,V> (f, shift);
}

TMPL Anfunc
operator / (const Anfunc& f, const Anfunc& g) {
  return binary_implicit_anfunc<div_op> (f, g);
}

ARITH_INT_SUGAR(TMPL,Anfunc,C);
//VAR_ARITH_INT_SUGAR(TMPL,Anfunc,C);
//TMPL inline Anfunc operator * (const Anfunc& x, int y) { return x * C(y); }
//TMPL inline Anfunc operator / (const Anfunc& x, int y) { return x / C(y); }
//TMPL inline Anfunc operator / (int x, const Anfunc& y) { return Anfunc(x)/y; }

/******************************************************************************
* Integration
******************************************************************************/

TMPL
class integrate_anfunc_rep: public Anfunc_rep {
protected:
  Anfunc f;
  C c;
public:
  integrate_anfunc_rep (const Anfunc& f2, const C& c2): f (f2), c (c2) {}
  Series Expand () const {
    return integrate_init (expand (f), c); }
  Error Tail_bound (nat order, Assume& a) const {
    Series s= expand (this->me (), order);
    return
      abs (s[order]) +
      (tail_bound (this->f, order, a) / (order+1)); }
  Anfunc Focus (const Scalar& z) const {
    return integrate_init (cast_mul<Anfunc > (z, focus (f, z)), c); }
  Anfunc Move (const Scalar& z) const {
    return integrate_init (move (f, z), eval (this->me (), z)); }
  Anfunc Derive () const {
    return f; }
};

TMPL Anfunc
integrate (const Anfunc& f) {
  return (Anfunc_rep*) new integrate_anfunc_rep<C,V> (f, C(0));
}

TMPL Anfunc
integrate_init (const Anfunc& f, const C& c) {
  return (Anfunc_rep*) new integrate_anfunc_rep<C,V> (f, c);
}

/******************************************************************************
* Elementary functions
******************************************************************************/

TMPL Anfunc
sqrt (const Anfunc& f) {
  return unary_implicit_anfunc<sqrt_op> (f);
}

TMPL Anfunc
sqrt_init (const Anfunc& f, const C& c) {
  return unary_implicit_anfunc<sqrt_op> (f, c);
}

TMPL Anfunc
exp (const Anfunc& f) {
  return unary_implicit_anfunc<exp_op> (f);
}

TMPL Anfunc
log (const Anfunc& f) {
  return unary_implicit_anfunc<log_op> (f);
}

TMPL Anfunc
log_init (const Anfunc& f, const C& c) {
  return unary_implicit_anfunc<log_op> (f, c);
}

TMPL Anfunc
pow (const Anfunc& f, const C& c) {
  return exp (c * log (f));
}

// NOTE: cos, sin and tan defined in linear_series.hpp

TMPL Anfunc
acos (const Anfunc& f) {
  return unary_implicit_anfunc<acos_op> (f);
}

TMPL Anfunc
acos_init (const Anfunc& f, const C& c) {
  return unary_implicit_anfunc<acos_op> (f, c);
}

TMPL Anfunc
asin (const Anfunc& f) {
  return unary_implicit_anfunc<asin_op> (f);
}

TMPL Anfunc
asin_init (const Anfunc& f, const C& c) {
  return unary_implicit_anfunc<asin_op> (f, c);
}

TMPL Anfunc
atan (const Anfunc& f) {
  return unary_implicit_anfunc<atan_op> (f);
}

TMPL Anfunc
atan_init (const Anfunc& f, const C& c) {
  return unary_implicit_anfunc<atan_op> (f, c);
}

#undef TMPL_DEF
#undef TMPL
#undef Anfunc
#undef Anfunc_rep
#undef Implicit_anfunc_rep
#undef Assume
#undef Scalar
#undef Series_variant
#undef Series
#undef Error
} // namespace mmx
#endif // __MMX_ANFUNC_HPP
