
/******************************************************************************
* MODULE     : operators.hpp
* DESCRIPTION: Templates and default implementations of several operations
* COPYRIGHT  : (C) 2003  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_OPERATORS_HPP
#define __MMX_OPERATORS_HPP
#include "type_props.hpp"
#include "mmx_identifiers.hpp"
namespace mmx {

inline nat get_series_cancel_order ();

/******************************************************************************
* Refer to standard C++ operators by another name
******************************************************************************/

#define TMPL template<typename T>

TMPL T neg_operator (const T& x) { return -x; }
TMPL T add_operator (const T& x, const T& y) { return x+y; }
TMPL T sub_operator (const T& x, const T& y) { return x-y; }
TMPL T mul_operator (const T& x, const T& y) { return x*y; }
TMPL T div_operator (const T& x, const T& y) { return x/y; }

TMPL bool equal_operator (const T& x, const T& y) { return x==y; }
TMPL bool unequal_operator (const T& x, const T& y) { return x!=y; }
TMPL bool less_operator (const T& x, const T& y) { return x<y; }
TMPL bool lesseq_operator (const T& x, const T& y) { return x<=y; }
TMPL bool gtr_operator (const T& x, const T& y) { return x>y; }
TMPL bool gtreq_operator (const T& x, const T& y) { return x>=y; }

#undef TMPL

/******************************************************************************
* Constructors, copying and constants
******************************************************************************/

struct copy_op {
  template<typename C> static inline C
  op (const C& x) { return x; }
  template<typename C> static inline C&
  in_place (C& x, const C& y) { return x= y; }
};

struct strict_copy_op {
  template<typename C> static inline C
  op (const C& x) { return copy (x); }
  static inline generic
  op (const generic& x) { return x; }
  template<typename C> static inline C&
  in_place (C& x, const C& y) { return x= copy (y); }
};

struct melt_op {
  static generic name () { return MMX_MELT; }
  template<typename C> static inline C
  op (const C& x) { return melt (x); }
};

struct fuse_op {
  static generic name () { return MMX_FUSE; }
  template<typename C> static inline C
  op (const C& x) { return fuse (x); }
};

struct swap_op {
  template<typename C> static inline void
  in_place (C& x, C& y) { C aux; x= y; y= aux; }
};

struct log2_op {
  template<typename C> static inline C
  op () { return Log2 (C); }
};

struct pi_op {
  template<typename C> static inline C
  op () { return Pi (C); }
};

struct euler_op {
  template<typename C> static inline C
  op () { return Euler (C); }
};

struct catalan_op {
  template<typename C> static inline C
  op () { return Catalan (C); }
};

struct imaginary_op {
  template<typename C> static inline C
  op () { return Imaginary (C); }
};

struct infinity_op {
  template<typename C> static inline C
  op () { return Infinity (C); }
};

struct plus_infinity_op {
  template<typename C> static inline C
  op () { return Plus_infinity (C); }
};

struct minus_infinity_op {
  template<typename C> static inline C
  op () { return Minus_infinity (C); }
};

/******************************************************************************
* Arithmetic operators
******************************************************************************/

struct neg_op {
  static generic name () { return MMX_MINUS; }
  template<typename C> static inline C
  op (const C& x) { return -x; }
  template<typename C> static inline C&
  in_place (C& x) { return x= -x; }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return -derive (x); }
};

struct add_op {
  static generic name () { return MMX_PLUS; }
  template<typename C> static inline C
  neutral () { return 0; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return x + y; }
  template<typename C> static inline C&
  in_place (C& x, const C& y) { return x += y; }
  template<typename C> static inline C
  diff_op (const C& me, const C& x, const C& y) {
    (void) me; return derive (x) + derive (y); }
};

struct sub_op {
  static generic name () { return MMX_MINUS; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return x - y; }
  template<typename C> static inline C&
  in_place (C& x, const C& y) { return x -= y; }
  template<typename C> static inline C
  diff_op (const C& me, const C& x, const C& y) {
    (void) me; return derive (x) - derive (y); }
};

template<bool left> struct sc_mul_op;

struct mul_op {
  typedef sc_mul_op<false> sc_left;
  typedef sc_mul_op<true> sc_right;
  static generic name () { return MMX_TIMES; }
  template<typename C> static inline C
  neutral () { return 1; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return x * y; }
  template<typename C> static inline C&
  in_place (C& x, const C& y) { return x *= y; }
  template<typename R> struct cast_helper {
    template<typename F1, typename F2> static inline R
    op (const F1& x, const F2& y) {
      return x * y; } };
  template<typename C> static inline C
  diff_op (const C& me, const C& x, const C& y) {
    (void) me; return derive (x)*y + x*derive (y); }
};

template<typename T,typename F1,typename F2> T
cast_mul (const F1& x, const F2& y) {
  return cast_binary<mul_op,T> (x, y);
}

template<bool left>
struct sc_mul_op {
  static generic name () { return MMX_TIMES; }
  template<typename C,typename S> static inline C
  op (const C& x, const S& y) { return x * y; }
  template<typename C,typename S> static inline C&
  in_place (C& x, const S& y) { return x *= y; }
  template<typename R> struct cast_helper {
    template<typename F1, typename F2> static inline R
    cast_op (const F1& x, const F2& y) {
      return cast_mul<R> (x, y); } };
  template<typename C,typename S> static inline C
  diff_op (const C& me, const C& x, const S& y) {
    (void) me; return derive (x) * y; }
};

STMPL
struct sc_mul_op<true> {
  static generic name () { return MMX_TIMES; }
  template<typename C,typename S> static inline C
  op (const C& y, const S& x) { return x * y; }
  template<typename R> struct cast_helper {
    template<typename F1, typename F2> static inline R
    cast_op (const F2& y, const F1& x) {
      return cast_mul<R> (x, y); } };
  template<typename C,typename S> static inline C
  diff_op (const C& me, const C& y, const S& x) {
    (void) me; return x * derive (y); }
};

template<bool left> struct sc_div_op;

struct div_op {
  typedef sc_div_op<false> sc_left;
  typedef sc_div_op<true> sc_right;
  static generic name () { return MMX_OVER; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return x / y; }
  template<typename C> static inline C&
  in_place (C& x, const C& y) { return x /= y; }
  template<typename R> struct cast_helper {
    template<typename F1, typename F2> static inline R
    op (const F1& x, const F2& y) {
      return x / y; } };
  template<typename C> static inline C
  diff_op (const C& me, const C& x, const C& y) {
    (void) me; return (derive (x)*y - x*derive (y)) / square (y); }
  static inline nat nr_init () { return 0; }
  template<typename C> static inline C
  def (const C& me, const C& f, const C& g) {
    if (g[0] == 0) {
      nat val;
      for (val=0; g[val]==0; val++)
	ASSERT_DOMAIN (val != get_series_cancel_order ()); // divide by zero?
      return shift_exp (f, -((int) val)) / shift_exp (g, -((int) val));
    }
    return (f - shift_exp (me * shift_exp (g, -1), 1)) / g[0]; }
};

template<typename T,typename F1,typename F2> T
cast_div (const F1& x, const F2& y) {
  return cast_binary<div_op,T> (x, y);
}

template<bool left>
struct sc_div_op {
  static generic name () { return MMX_OVER; }
  template<typename C,typename S> static inline C
  op (const C& x, const S& y) { return x / y; }
  template<typename C,typename S> static inline C&
  in_place (C& x, const S& y) { return x /= y; }
  template<typename R> struct cast_helper {
    template<typename F1, typename F2> static inline R
    cast_op (const F1& x, const F2& y) {
      return cast_div<R> (x, y); } };
  template<typename C,typename S> static inline C
  diff_op (const C& me, const C& x, const S& y) {
    (void) me; return derive (x) / y; }
};

STMPL
struct sc_div_op<true> {
  static generic name () { return MMX_OVER; }
  template<typename C,typename S> static inline C
  op (const C& y, const S& x) { return x / y; }
  template<typename R> struct cast_helper {
    template<typename F1, typename F2> static inline R
    cast_op (const F2& y, const F1& x) {
      return cast_div<R> (y, x); } };
  template<typename C,typename S> static inline C
  diff_op (const C& me, const C& y, const S& x) {
    (void) me; return -x * derive (y) / square (y); }
};

struct quo_op {
  static generic name () { return MMX_DIV; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return quo (x, y); }
  static inline nat nr_init () { return 0; }
  template<typename C> static inline C
  def (const C& me, const C& f, const C& g) {
    if (g[0] == 0) {
      nat val;
      for (val=0; g[val]==0; val++)
	if (val == get_series_cancel_order ()) return 0;
      return quo (shift_exp (f, -((int) val)), shift_exp (g, -((int) val)));
    }
    return quo (f - shift_exp (me * shift_exp (g, -1), 1), g[0]); }
};

template<bool left>
struct sc_quo_op {
  static generic name () { return MMX_DIV; }
  template<typename C,typename S> static inline C
  op (const C& x, const S& y) { return quo (x, y); }
};

struct rem_op {
  static generic name () { return MMX_MOD; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return rem (x, y); }
};

template<bool left>
struct sc_rem_op {
  static generic name () { return MMX_MOD; }
  template<typename C,typename S> static inline C
  op (const C& x, const S& y) { return rem (x, y); }
};

template<bool left>
struct lsh_op {
  template<typename C,typename S> static inline C
  op (const C& x, const S& sh) { return x << sh; }
  template<typename C,typename S> static inline C&
  in_place (C& x, const S& sh) { return x <<= sh; }
};

template<bool left>
struct rsh_op {
  template<typename C,typename S> static inline C
  op (const C& x, const S& sh) { return x >> sh; }
  template<typename C,typename S> static inline C&
  in_place (C& x, const S& sh) { return x >>= sh; }
};

struct or_op {
  template<typename C> static inline C
  neutral () { return false; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return x | y; }
  template<typename C> static inline C&
  in_place (C& x, const C& y) { return x |= y; }
};

struct and_op {
  template<typename C> static inline C
  neutral () { return true; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return x & y; }
  template<typename C> static inline C&
  in_place (C& x, const C& y) { return x &= y; }
};

/******************************************************************************
* Predicates
******************************************************************************/

struct hard_eq_op { // hard equality in memory
  static generic name () { return MMX_HARD_EQ; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return hard_eq (x, y); }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return hard_neq (x, y); }
};

struct hard_neq_op { // hard inequality in memory
  static generic name () { return MMX_HARD_NEQ; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return hard_neq (x, y); }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return hard_eq (x, y); }
};

struct eq_op { // syntactic equality
  static generic name () { return MMX_EQ; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return eq (x, y); }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return neq (x, y); }
};

struct neq_op { // syntactic inequality
  static generic name () { return MMX_NEQ; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return neq (x, y); }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return eq (x, y); }
};

struct equal_op {
  static generic name () { return MMX_EQUAL; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return x == y; }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return x != y; }
};

struct unequal_op {
  static generic name () { return MMX_UNEQUAL; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return x != y; }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return x == y; }
};

struct less_op {
  static generic name () { return MMX_LESS; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return x < y; }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return x >= y; }
};

struct lesseq_op {
  static generic name () { return MMX_LESSEQ; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return x <= y; }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return x > y; }
};

struct gtr_op {
  static generic name () { return MMX_GTR; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return x > y; }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return x <= y; }
};

struct gtreq_op {
  static generic name () { return MMX_GTREQ; }
  template<typename C> static inline bool
  op (const C& x, const C& y) { return x >= y; }
  template<typename C> static inline bool
  not_op (const C& x, const C& y) { return x < y; }
};

/******************************************************************************
* Ordering related operators
******************************************************************************/

struct min_op {
  template<typename C> static inline C
  neutral () { return Plus_infinity (C); }
  template<typename C> static inline C
  op (const C& x, const C& y) { return min (x, y); }
  template<typename C> static inline C&
  in_place (C& x, const C& y) { return x= min (x, y); }
};

struct max_op {
  template<typename C> static inline C
  neutral () { return Minus_infinity (C); }
  template<typename C> static inline C
  op (const C& x, const C& y) { return max (x, y); }
  template<typename C> static inline C&
  in_place (C& x, const C& y) { return x= max (x, y); }
};

/******************************************************************************
* Operators on complex number like types
******************************************************************************/

struct gaussian_op {
  template<typename C> static inline Complex_type(C)
  op (const C& x, const C& y) { return gaussian (x, y); }
  static inline generic
  op (const generic& x, const generic& y) { return gaussian (x, y); }
};

struct polar_op {
  template<typename C> static inline Complex_type(C)
  op (const C& x, const C& y) { return polar (x, y); }
  static inline generic
  op (const generic& x, const generic& y) { return polar (x, y); }
};

struct Re_op {
  template<typename C> static inline Real_type(C)
  op (const C& z) { return Re (z); }
  static inline generic
  op (const generic& z) { return Re (z); }
};

struct Im_op {
  template<typename C> static inline Real_type(C)
  op (const C& z) { return Im (z); }
  static inline generic
  op (const generic& z) { return Im (z); }
};

template<typename T,typename F> T cast_abs (const F& x);

struct abs_op {
  template<typename C> static inline Abs_type(C)
  op (const C& z) { return abs (z); }
  static inline generic
  op (const generic& z) { return abs (z); }
  template<typename R> struct cast_helper {
    template<typename C> static inline R
    op (const C& x) { return abs (x); }
    template<typename C> static inline R
    cast_op (const C& x) { return cast_abs<R> (x); } };
};

template<typename T,typename F> T
cast_abs (const F& x) {
  return cast_unary<abs_op,T> (x);
}

struct arg_op {
  template<typename C> static inline Real_type(C)
  op (const C& z) { return arg (z); }
  static inline generic
  op (const generic& z) { return arg (z); }
};

struct conj_op {
  template<typename C> static inline C
  op (const C& z) { return conj (z); }
};

/******************************************************************************
* Elementary functions
******************************************************************************/

struct sqrt_op {
  static generic name () { return MMX_SQRT; }
  template<typename C> static inline C
  op (const C& x) { return sqrt (x); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return sqrt_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    return derive (x) / (2 * me); }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    if (f[0] == 0) {
      nat val= 0;
      for (val=0; val<get_series_cancel_order() && f[val]==0; val++);
      ASSERT_DOMAIN (val<get_series_cancel_order() && (val&1) == 0);
      return shift_exp (sqrt (shift_exp (f, -((int) val))), val/2);
    }
    return integrate (derive (f) / (2*me)); }
};

struct exp_op {
  static generic name () { return MMX_EXP; }
  template<typename C> static inline C
  op (const C& x) { return exp (x); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { (void) i; return exp (x); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    return derive (x) * me; }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

struct log_op {
  static generic name () { return MMX_LOG; }
  template<typename C> static inline C
  op (const C& x) { return log (x); }
  template<typename C,typename I> static inline C
  op_init (const C& x, const I& i) { return log_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) / x; }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

struct pow_op {
  static generic name () { return MMX_POWER; }
  template<typename C> static inline C
  op (const C& x, const C& y) { return pow (x, y); }
};

struct trig_op {
  template<typename C> static inline C
  op (const C& x) { return trig (x); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { (void) i; return trig (x); }
  static inline generic
  op (const generic& x) { return xsqtuple (comma (cos (x), sin (x))); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x);
  static inline nat nr_init () { return 1; }
  template<typename V,typename C> static inline V
  def (const V& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

struct cos_op {
  static generic name () { return MMX_COS; }
  template<typename C> static inline C
  op (const C& x) { return cos (x); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { (void) i; return cos (x); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return -derive (x) * sin (x); }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    (void) me; return cos_sin (f) [0]; }
};

struct sin_op {
  static generic name () { return MMX_SIN; }
  template<typename C> static inline C
  op (const C& x) { return sin (x); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { (void) i; return sin (x); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) * cos (x); }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    (void) me; return cos_sin (f) [1]; }
};

struct tan_op {
  static generic name () { return MMX_TAN; }
  template<typename C> static inline C
  op (const C& x) { return tan (x); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { (void) i; return tan (x); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) / square (cos (x)); }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f);
};

struct acos_op {
  static generic name () { return MMX_ARCCOS; }
  template<typename C> static inline C
  op (const C& x) { return acos (x); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return acos_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return -derive (x) / sqrt (1 - square (x)); }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

struct asin_op {
  static generic name () { return MMX_ARCSIN; }
  template<typename C> static inline C
  op (const C& x) { return asin (x); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return asin_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) / sqrt (1 - square (x)); }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

struct atan_op {
  static generic name () { return MMX_ARCTAN; }
  template<typename C> static inline C
  op (const C& x) { return atan (x); }
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return atan_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x) {
    (void) me; return derive (x) / (1 - square (x)); }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

/******************************************************************************
* Calculus
******************************************************************************/

struct derive_op {
  template<typename C> static inline C
  op (const C& x) { return derive (x); }
  template<typename C,typename V> static inline C
  op (const C& x, const V& v) { return derive (x, v); }
};

struct xderive_op {
  template<typename C> static inline C
  op (const C& x) { return xderive (x); }
  template<typename C,typename V> static inline C
  op (const C& x, const V& v) { return xderive (x, v); }
};

struct integrate_op {
  template<typename C> static inline C
  op (const C& x) { return integrate (x); }
  template<typename C,typename V> static inline C
  op (const C& x, const V& v) { return integrate (x, v); }
  template<typename C,typename I> static inline C
  op_init (const C& x, const I& i) { return integrate_init (x, i); }
  template<typename C,typename V,typename I> static inline C
  op_init (const C& x, const V& v, const I& i) {
    return integrate_init (x, v, i); }
};

struct solve_vector_lde_op {
  template<typename C> static inline C
  op (const C& x) { assert (false); return x; } // must initialize
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return solve_vector_lde_init (x, i); }
  static inline generic op_init (const generic& x, const generic& i) { 
    return solve_lde_init (x, i); }
  template<typename C> static inline C
  diff_op (const C& me, const C& x);
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C
  def (const C& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

struct solve_matrix_lde_op {
  template<typename C> static inline C
  op (const C& x) { assert (false); return x; } // must initialize
  template<typename C, typename I> static inline C
  op_init (const C& x, const I& i) { return solve_matrix_lde_init (x, i); }
  static inline generic op_init (const generic& x, const generic& i) { 
    return solve_lde_init (x, i); }
  template<typename C> static inline C diff_op (const C& me, const C& x) {
    return x * me; }
  static inline nat nr_init () { return 1; }
  template<typename C> static inline C def (const C& me, const C& f) {
    return integrate (diff_op (me, f)); }
};

} // namespace mmx
#endif // __MMX_OPERATORS_HPP
