
/******************************************************************************
* MODULE     : routine.cpp
* DESCRIPTION: Abstract routines on generic expressions
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "routine.hpp"
#include "generic_internal.hpp"
#include "literal.hpp"
#include "glue.hpp"
#include "environment.hpp"
#include "exception.hpp"
#include "tuple.hpp"
namespace mmx {

/******************************************************************************
* Routines for debugging
******************************************************************************/

array<generic>
type_name (const array<nat>& ids) {
  nat i, n= N(ids);
  array<generic> ret= array_fill<generic> (n);
  for (i=0; i<n; i++)
    ret(i)= type_name (ids[i]);
  return ret;
}

/******************************************************************************
* Generic routines
******************************************************************************/

generic
routine_rep::apply () const {
  return xapply (gen (name));
}

generic
routine_rep::apply (const generic& g1) const {
  return xapply (gen (name, g1));
}

generic
routine_rep::apply (const generic& g1, const generic& g2) const {
  return xapply (gen (name, g1, g2));
}

generic
routine_rep::xapply (const generic& x) const {
  cerr << "name= " << name << "\n";
  cerr << "x   = " << x << "\n";
  assert (false); // not applicable
  return generic ();
}

generic
routine_rep::apply (const array<generic>& v) const {
  switch (N(v)) {
  case 0:
    return apply ();
  case 1:
    return apply (v[0]);
  case 2:
    return apply (v[0], v[1]);
  default:
    return xapply (gen (name, v));
  }
}

array<nat>
routine_rep::signature () const {
  assert (false); // not applicable
  return array<nat> ();
}

void
routine_rep::overload (const routine& fun) const {
  assert (false); // not applicable
}

/******************************************************************************
* Identity routines
******************************************************************************/

class identity_routine_rep: public routine_rep {
  array<nat> sig;
public:
  identity_routine_rep (const array<nat>& sig2):
    routine_rep ("identity"), sig (sig2) {}
  generic apply (const generic& x) const { return x; }
  generic xapply (const generic& g) const {
    assert (N(g)==2); return apply (g[1]); }
  array<nat> signature () const { return sig; }
};

routine
identity_routine (const array<nat>& sig) {
  // WARNING: we do allow for "fake identity functions" with sig[0] != sig[1]
  // This is useful for fast conversions to generic
  assert (N(sig) == 2);
  return new identity_routine_rep (sig);
}

/******************************************************************************
* Composition of routines
******************************************************************************/

class composed_routine_rep: public routine_rep {
  routine fun;
  array<routine> args;
  array<nat> sig;
public:
  composed_routine_rep (const routine& fun2, const array<routine>& args2):
    routine_rep (compose (fun2->name, flatten (args2))),
    fun (fun2), args (args2)
  {
    sig= copy (fun->signature ());
    assert (N(sig) == N(args)+1);
    for (nat i=0; i<N(args); i++) {
      array<nat> sub_sig= args[i]->signature ();
      assert (N(sub_sig) == 2);
      assert (sig[i+1] == sub_sig[0] || sig[i+1] == 0);
      sig(i+1)= sub_sig[1];
    }
  }
  generic apply (const generic& x) const {
    assert (N(args) == 1);
    return fun->apply (args[0]->apply (x));
  }
  generic apply (const generic& x, const generic& y) const {
    assert (N(args) == 2);
    return fun->apply (args[0]->apply (x), args[1]->apply (y));
  }
  generic xapply (const generic& x) const {
    nat i, n= N(x)-1;
    assert (N(args) == n);
    array<generic> v= array_fill<generic> (n);
    for (i=0; i<n; i++)
      v(i)= args[i]->apply (x[i+1]);
    return fun->xapply (gen (x[0], v));
  }
  array<nat> signature () const { return sig; }
};

routine
compose (const routine& fun, const array<routine>& args) {
  // WARNING: the signature of the result is not necessarily correct
  // when simplifying with "fake identity functions", but that does not
  // matter for the application to overloading
  if (eq (fun->name, generic ("identity")) && N(args) == 1)
    return args [0];
  for (nat i=0; i<N(args); i++)
    if (neq (args[i]->name, generic ("identity")))
      return new composed_routine_rep (fun, args);
  return fun;
}

/******************************************************************************
* Exception routines
******************************************************************************/

class exception_routine_rep: public routine_rep {
public:
  exception_routine_rep (): routine_rep ("error") {}
  generic xapply (const generic& x) const {
    for (nat i=1; i<N(x); i++)
      if (is<exception> (x[i])) return x[i];
    return x; }
  array<nat> signature () const { return array<nat> (); }
};

routine
exception_routine () {
  return new exception_routine_rep ();
}

/******************************************************************************
* Flatten out tuple arguments in one big tuple
******************************************************************************/

generic
equalize (const generic& name, const generic& x) {
  array<generic> v= array<generic> (name);
  for (nat i=1; i<N(x); i++) {
    if (is_tuple_type (type (x[i]))) {
      array<generic> w= compound_to_array (*as<tuple<generic> > (x[i]));
      v << cdr (w);
    }
    else v << x[i];
  }
  return array_to_compound (v);
}

class equalize_tuple_routine_rep: public routine_rep {
  routine fun;
  array<nat> sig;
public:
  equalize_tuple_routine_rep (const routine& fun2, const array<nat>& sig2):
    routine_rep (gen ("equalize_tuple", fun2->name)),
    fun (fun2), sig (sig2) {}
  generic xapply (const generic& g) const {
    return fun->xapply (equalize (MMX_TUPLE, g)); }
  array<nat> signature () const { return sig; }
};

routine
equalize_tuple_routine (const routine& fun, const array<nat>& sig) {
  return new equalize_tuple_routine_rep (fun, sig);
}

/******************************************************************************
* Apply routines which operate on tuples
******************************************************************************/

class via_tuple_routine_rep: public routine_rep {
  routine fun;
  array<nat> sig;
  nat n;
public:
  via_tuple_routine_rep (const routine& fun2, const array<nat>& sig2, nat n2):
    routine_rep (gen ("via_tuple", fun2->name)),
    fun (fun2), sig (sig2), n (n2) {}
  generic xapply (const generic& g) const {
    array<generic> v= compound_to_array (g);
    array<generic> w= range (v, 0, n-1);
    generic t= gen (MMX_TUPLE, range (v, n-1, N(v)));
    w << as<generic> (tuple<generic> (t));
    return fun->xapply (array_to_compound (w)); }
  array<nat> signature () const { return sig; }
};

routine
via_tuple_routine (const routine& fun, const array<nat>& sig, nat n) {
  return new via_tuple_routine_rep (fun, sig, n);
}

/******************************************************************************
* Specialization of Alias Generic to Generic_alias T for some T
******************************************************************************/

class specialize_alias_routine_rep: public routine_rep {
  routine fun;
  array<nat> sig;
public:
  specialize_alias_routine_rep (const routine& fun2, const array<nat>& sig2):
    routine_rep (gen ("specialize_alias", fun2->name)),
    fun (fun2), sig (sig2) {}
  generic xapply (const generic& g) const {
    nat i, n= N(g);
    array<generic> v= array_fill<generic> (n);
    v(0)= MMX_TUPLE;
    for (i=1; i<n; i++)
      if (type (g[i]) == type_id<alias<generic> > ())
	v(i)= specialize_alias (g[i]);
      else v(i)= g[i];
    return fun->xapply (array_to_compound (v)); }
  array<nat> signature () const { return sig; }
};

routine
specialize_alias_routine (const routine& fun, const array<nat>& sig) {
  return new specialize_alias_routine_rep (fun, sig);
}

/******************************************************************************
* Default routines
******************************************************************************/

class default_routine_rep: public routine_rep {
public:
  default_routine_rep (const generic& name): routine_rep (name) {}
  generic apply () const { return gen (name); }
  generic apply (const generic& x1) const {
    if (is_tuple_type (type (x1)))
      return xapply (gen (name, x1));
    if (is_tuple_type (type (x1))) return xapply (gen (name, x1));
    return gen (name, x1); }
  generic apply (const generic& x1, const generic& x2) const {
    if (is_tuple_type (type (x1)) || is_tuple_type (type (x2)))
      return xapply (gen (name, x1, x2));
    return gen (name, x1, x2); }
  generic xapply (const generic& x) const {
    for (nat i=1; i<N(x); i++)
      if (is_tuple_type (type (x[i])))
	return equalize (name, x);
    return cons (name, cdr (x)); }
  array<nat> signature () const { return array<nat> (); }
};

routine
default_routine (const generic& name) {
  return new default_routine_rep (name);
}

/******************************************************************************
* Overloaded generic routines
******************************************************************************/

static array<nat>
type (const array<generic>& args) {
  nat i, n= N(args);
  array<nat> r= array_fill<nat> ((nat) 0, n);
  for (i=0; i<n; i++)
    r(i)= type (args[i]);
  return r;
}

static array<nat>
xtype (const generic& g) {
  nat i, n= N(g)-1;
  array<nat> r= array_fill<nat> ((nat) 0, n);
  for (i=0; i<n; i++)
    r(i)= type (g[i+1]);
  return r;
}

static nat
conversion_penalty (const environment& env, nat id1, nat id2) {
  nat penalty;
  generic r= env->get_converter (id1, id2, penalty);
  //cout << "  Convert " << id1 << ", " << id2 << " -> " << r << "\n";
  assert (is<routine> (r));
  return penalty;
}

static array<nat>
untuple (const array<nat>& ids, nat total) {
  nat i, n= N(ids), tupid= tuple_to_scalar (ids[n-1]);
  array<nat> r= array_fill<nat> ((nat) 0, total);
  for (i=0; i<n-1; i++)
    r(i)= ids[i];
  for (; i<total; i++)
    r(i)= tupid;
  return r;
}

static nat
conversion_penalty (const environment& env,
		    const array<nat>& ids1, const array<nat>& ids2)
{
  if (N(ids1) < N(ids2)-1) return PENALTY_INVALID;
  if (is_tuple_type (ids2[N(ids2)-1]))
    return conversion_penalty (env, ids1, untuple (ids2, N(ids1)));
  if (N(ids1) != N(ids2)) return PENALTY_INVALID;
  nat i, n= N(ids1), penalty= 0;
  for (i=1; i<n; i++)
    penalty= max (penalty, conversion_penalty (env, ids1[i], ids2[i]));
  return penalty;
}

static routine
build (const environment& env, const routine& fun,
       const array<nat>& ids1, const array<nat>& ids2)
{
  if (is_tuple_type (ids2[N(ids2)-1])) {
    array<nat> ids3= untuple (ids2, N(ids1));
    routine vtfun= via_tuple_routine (fun, ids3, N(ids2));
    return build (env, vtfun, ids1, ids3);
  }
  nat i, n= N(ids1) - 1;
  array<routine> v= array_fill<routine> (n);
  for (i=0; i<n; i++) {
    nat penalty;
    generic r= env->get_converter (ids1[i+1], ids2[i+1], penalty);
    assert (is<routine> (r));
    v(i)= as<routine> (r);
  }
  return compose (fun, v);
}

class overloaded_routine_rep: public routine_rep {
  array<routine>     funs;
  environment        env;    // environment with applicable converters
  nat                serial; // time stamp for correctness of cache
  routine            nullary;
  table<routine,nat> unary;
  table<routine,nat> binary;
  table<routine,array<nat> > n_ary;
  routine            fall_back;

public:
  overloaded_routine_rep (const generic& name, const environment& env2):
    routine_rep (name),
    env (env2),
    serial (env->serial),
    nullary (default_routine (name)),
    fall_back (default_routine (name)) {}
  void invalidate () const {
    overloaded_routine_rep* me=
      const_cast<overloaded_routine_rep*> (this);
    me->unary = table<routine,nat> ();
    me->binary= table<routine,nat> ();
    me->n_ary = table<routine,array<nat> > ();
    me->serial= env->serial;
  }
  inline void up_to_date () const {
    if (!nil (env->next) && env->next_serial != env->next->serial)
      env->ensure_up_to_date ();
    if (serial != env->serial) invalidate ();
  }
  routine resolve (const array<generic>& args) const {
    overloaded_routine_rep* me=
      const_cast<overloaded_routine_rep*> (this);
    routine best;
    array<nat> ids= cons<nat> (0, type (args));
    bool tuple_args= false, genalias_args= false, exc_args= false;
    for (nat i=1; i<N(ids); i++) {
      exc_args= exc_args || ids[i] == type_id<exception> ();
      tuple_args= tuple_args || is_tuple_type (ids[i]);
      genalias_args= genalias_args || ids[i] == type_id<alias<generic> > ();
    }
    if (exc_args) best= exception_routine ();
    else if (tuple_args)
      best= equalize_tuple_routine (routine (me, true), ids);
    else if (genalias_args)
      best= specialize_alias_routine (routine (me, true), ids);
    else {
      array<nat> best_ids;
      nat best_pen= PENALTY_INVALID;
      for (nat i=0; i<N(funs); i++) {
	array<nat> fun_ids= funs[i]->signature ();
	nat pen= conversion_penalty (env, ids, fun_ids);
	//if (!using_simple ())
	//cout << "Try " << name << ", " << type_name (fun_ids)
	//<< " on " << type_name (ids) << "\n";
	if (pen <= best_pen && pen < PENALTY_INVALID) {
	  if (pen < best_pen ||
	      conversion_penalty (env, fun_ids, best_ids) <
	      conversion_penalty (env, best_ids, fun_ids))
	    {
	      //if (!using_simple ())
	      //cout << "  OK, and better (" << pen << ")\n";
	      best= build (env, funs[i], ids, fun_ids);
	      best_ids= fun_ids;
	      best_pen= pen;
	    }
	  //else if (!using_simple ())
	  //cout << "  OK, but less good (" << pen << ")\n";
	}
      }
    }
    if (nil (best)) best= fall_back;
    if (N(ids) == 2) me->unary (ids[1])= best;
    else if (N(ids) == 3) me->binary (binary_id (ids[1], ids[2]))= best;
    if (N(ids) <= 7) me->n_ary (cdr (ids))= best;
    return best;
  }
  generic apply () const {
    return nullary->apply ();
  }
  generic apply (const generic& x1) const {
    up_to_date ();
    routine fun= unary[type (x1)];
    if (nil (fun)) fun= resolve (array<generic> (x1));
    return fun->apply (x1);
  }
  generic apply (const generic& x1, const generic& x2) const {
    up_to_date ();
    nat id1= type (x1), id2= type (x2);
    routine fun= binary[binary_id (id1, id2)];
    if (nil (fun)) fun= resolve (array<generic> (x1, x2));
    return fun->apply (x1, x2);
  }
  generic xapply (const generic& x) const {
    up_to_date ();
    routine fun= n_ary[xtype (x)];
    if (nil (fun)) fun= resolve (compound_to_array (cdr (x)));
    return fun->xapply (x);
  }
  void overload (const routine& fun) const {
    overloaded_routine_rep* me=
      const_cast<overloaded_routine_rep*> (this);
    array<nat> ids= fun->signature ();
    /*
    if (!using_simple () && eq (fun->name, MMX_TIMES)) {
      cout << fun << "\t: " << type_name (ids) << "\n";
      //cout << "\t: " << env << "\n";
    }
    */
    if (N(ids) == 0) me->fall_back= fun;
    else if (N(ids) == 1) me->nullary= fun;
    else {
      nat i, n= N(funs);
      for (i=0; i<n; i++) {
	array<nat> ids2= funs[i]->signature ();
	if (cdr (ids) == cdr (ids2)) {
	  me->funs(i)= fun;
	  break;
	}
      }
      if (i==n) me->funs << fun;
    }
    me->invalidate ();
  }
  array<nat> signature () const { return array<nat> (); }
};

routine
overloaded_routine (const generic& name, const environment& env) {
  return new overloaded_routine_rep (name, env);
}

} // namespace mmx
