
/******************************************************************************
* MODULE     : commutative_glue.cpp
* DESCRIPTION: Standard rules for commutative arithmetic
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "evaluator.hpp"
#include "category_glue.hpp"
#include "rational.hpp"
#include "big_sum.hpp"
#include "big_product.hpp"
#include "tuple.hpp"
namespace mmx {
#define Table table<generic,generic>
#define Iterator iterator<generic>

/******************************************************************************
* Big sums
******************************************************************************/

generic sum_neg (const big_sum& s) {
  return table_to_sum (-(*s)); }
generic sum_add (const big_sum& s1, const big_sum& s2) {
  return table_to_sum ((*s1) + (*s2)); }
generic sum_sub (const big_sum& s1, const big_sum& s2) {
  return table_to_sum ((*s1) - (*s2)); }
generic commutative_neg (const generic& s) {
  return table_to_sum (-sum_to_table (s)); }
generic commutative_add (const generic& s1, const generic& s2) {
  return table_to_sum (sum_to_table (s1) + sum_to_table (s2)); }
generic commutative_sub (const generic& s1, const generic& s2) {
  return table_to_sum (sum_to_table (s1) - sum_to_table (s2)); }

generic
commutative_sum (const tuple<generic>& x) {
  Table r (0);
  nat i, n= N (x);
  for (i=1; i<n; i++) {
    Table t= sum_to_table (x[i]);
    for (Iterator it= entries (t); busy (it); ++it)
      r(*it) += t[*it];
  }
  return table_to_sum (r);
}

/******************************************************************************
* Big products
******************************************************************************/

generic product_mul (const big_product& p1, const big_product& p2) {
  return table_to_product ((*p1) + (*p2)); }
generic product_div (const big_product& p1, const big_product& p2) {
  return table_to_product ((*p1) - (*p2)); }
generic mul_term (const generic& p1, const generic& p2) {
  return table_to_product (product_to_table (p1) + product_to_table (p2)); }
generic div_term (const generic& p1, const generic& p2) {
  return table_to_product (product_to_table (p1) - product_to_table (p2)); }

static Table
mul (const Table& t1, const Table& t2) {
  Table r (0);
  for (Iterator it1= entries (t1); busy (it1); ++it1)
    for (Iterator it2= entries (t2); busy (it2); ++it2) {
      generic x= mul_term (*it1, *it2);
      r(x) += t1[*it1] * t2[*it2];
    }
  return r;
}

generic
commutative_mul (const generic& x1, const generic& x2) {
  if (x1 == 0) return 0;
  if (x2 == 0) return 0;
  if (x1 == 1) return x2;
  if (x2 == 1) return x1;
  if (scalar_expand()) {
    if (is_scalar (x1))
      return table_to_sum (x1 * sum_to_table (x2));
    if (is_scalar (x2))
      return table_to_sum (x2 * sum_to_table (x1));
  }
  if (central_expand())
    return table_to_sum (mul (sum_to_table (x1), sum_to_table (x2)));
  return mul_term (x1, x2);
}

static Table
div (const Table& t, const generic& sc, const generic& x) {
  Table r (0);
  for (Iterator it= entries (t); busy (it); ++it) {
    generic y= div_term (*it, x);
    r(y)= div_term (t[*it], sc);
  }
  return r;
}

generic
commutative_div (const generic& x1, const generic& x2) {
  if (x1 == 0) return 0;
  if (x2 == 1) return x1;
  ASSERT (x2 != 0, "division by zero");
  if (scalar_expand())
    if (is_scalar (x2))
      return table_to_sum (sum_to_table (x1) / x2);
  if (central_expand()) {
    Table t= sum_to_table (x2);
    if (N(t) == 1) {
      generic y = *entries (t);
      generic sc= t[y];
      return table_to_sum (div (sum_to_table (x1), sc, y));
    }
    return table_to_sum (div (sum_to_table (x1), generic (1), x2));
  }
  else return div_term (x1, x2);
}

generic
commutative_pow (const generic& x1, const generic& x2) {
  if (x1 == 0) return x1;
  if (x1 == 1) return x1;
  if (x2 == 0) return 1;
  if (x2 == 1) return x1;
  if (x2 == -1 && is_scalar (x1)) return 1 / x1;
  Table t= product_to_table (x1);
  if (central_power_expand() || (scalar_power_expand() && is_scalar (x2)))
    return table_to_product (x2 * t);
  if (N(t) == 1)
    if (central_power_multiply() ||
	(scalar_power_multiply() && is_scalar (x2)))
      return table_to_product (x2 * t);
  Table r (0);
  r(x1)= x2;
  return as<generic> (big_product (r));

  // not commutative
  //if (is_func (x1, MMX_POWER, 2))
  //if (general_power_multiply() ||
  //(scalar_power_multiply() && is_scalar (x2)))
  //return commutative_pow (x1[1], x1[2] * x2);
  //return apply (MMX_POWER, x1, x2);
}

generic
commutative_product (const tuple<generic>& x) {
  generic prod= 1;
  nat i, j, n= N (x);
  for (i=n-1; i!=0; i=j) {
    Table r (0);
    for (j=i; j!=0; j--) {
      if (x[j] == 0) return 0;
      if (!is_central (x[j])) break;
      nat terms= N (sum_to_table (x[j]));
      if (terms > 1 && central_expand()) break;
      Table t= product_to_table (x[j]);
      for (Iterator it= entries (t); busy (it); ++it)
	r(*it) += t[*it];
    }
    prod= table_to_product (r) * prod;
    if (j>0) prod= x[j--] * prod;
  }
  return prod;
}

/******************************************************************************
* Interface
******************************************************************************/

void
define_commutative () {
  define_type<always,big_sum > ("Sum");
  define_type<always,big_product > ("Product");
  define_field<always,rational> ();
  define_ordering<always,rational> ();
  define<always> (MMX_MINUS, sum_neg);
  define<always> (MMX_PLUS , sum_add);
  define<always> (MMX_MINUS, sum_sub);
  define<always> (MMX_MINUS, commutative_neg);
  define<always> (MMX_PLUS , commutative_add);
  define<always> (MMX_MINUS, commutative_sub);
  define<always> (MMX_PLUS , commutative_sum);
  define<always> (MMX_TIMES, product_mul);
  define<always> (MMX_OVER , product_div);
  define<always> (MMX_TIMES, commutative_mul);
  define<always> (MMX_OVER , commutative_div);
  define<always> (MMX_POWER, commutative_pow);
  define<always> (MMX_TIMES, commutative_product);
}

#undef Table
#undef Iterator
} // namespace mmx
