
/******************************************************************************
* MODULE     : elementary_glue.cpp
* DESCRIPTION: Evaluation of elementary functions on generic expressions
* COPYRIGHT  : (C) 2006  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include "evaluator.hpp"
#include "glue.hpp"
#include "rational.hpp"
#include "literal.hpp"
#include "compound.hpp"
#include "big_sum.hpp"
#include "big_product.hpp"
#include "generic_internal.hpp"
namespace mmx {
#define Table table<generic,generic>
#define Array array<generic>
#define Iterator iterator<generic>

/******************************************************************************
* Exp-log functions
******************************************************************************/

generic
calculus_sqrt (const generic& x) {
  return pow (x, generic (1) / generic (2));
}

generic
calculus_exp (const generic& x) {
  return pow (generic (MMX_E), x);
}

generic
calculus_log (const generic& x) {
  if (x == 1) return 0;
  if (x == generic (MMX_E)) return 1;
  if (is<big_sum> (x)) {
    Table t= sum_to_table (x);
    if (N(t) == 1) {
      generic y = *entries (t);
      generic sc= t[y];
      return log (sc) + log (y);
    }
    else return gen (MMX_LOG, x);
  }
  if (is<big_product> (x)) {
    Table t= product_to_table (x);
    Array v;
    for (Iterator it= entries (t); busy (it); ++it)
      v << t[*it] * log (*it);
    return big_add (v);
  }
  return gen (MMX_LOG, x);
}

/******************************************************************************
* Trigonometric functions
******************************************************************************/

generic
calculus_cos (const generic& x) {
  if (x == 0) return 1;
  if (is_func (x, MMX_ARCCOS, 1)) return x[1];
  return gen (MMX_COS, x);
}

generic
calculus_sin (const generic& x) {
  if (x == 0) return 0;
  if (is_func (x, MMX_ARCSIN, 1)) return x[1];
  return gen (MMX_SIN, x);
}

generic
calculus_tan (const generic& x) {
  if (x == 0) return 0;
  if (is_func (x, MMX_ARCTAN, 1)) return x[1];
  return gen (MMX_TAN, x);
}

generic
calculus_acos (const generic& x) {
  if (x == 1) return 0;
  if (is_func (x, MMX_COS, 1)) return x[1];
  return gen (MMX_ARCCOS, x);
}

generic
calculus_asin (const generic& x) {
  if (x == 0) return 0;
  if (is_func (x, MMX_SIN, 1)) return x[1];
  return gen (MMX_ARCSIN, x);
}

generic
calculus_atan (const generic& x) {
  if (x == 0) return 0;
  if (is_func (x, MMX_TAN, 1)) return x[1];
  return gen (MMX_ARCTAN, x);
}

/******************************************************************************
* Interface
******************************************************************************/

void
define_elementary () {
  define<always> (MMX_SQRT, calculus_sqrt);
  define<always> (MMX_EXP, calculus_exp);
  define<always> (MMX_LOG, calculus_log);
  define<always> (MMX_COS, calculus_cos);
  define<always> (MMX_SIN, calculus_sin);
  define<always> (MMX_TAN, calculus_tan);
  define<always> (MMX_ARCCOS, calculus_acos);
  define<always> (MMX_ARCSIN, calculus_asin);
  define<always> (MMX_ARCTAN, calculus_atan);
}

#undef Table
#undef Array
#undef Iterator
} // namespace mmx
